#include <stdio.h>
#include <time.h>
#include <utime.h>
#include "t3f/t3f.h"

#include "body.h"
#include "pcx.h"

T3F_ANIMATION * vgolf_bitmap_to_animation(ALLEGRO_BITMAP * bp)
{
	T3F_ANIMATION * ap = NULL;
	if(bp)
	{
		ap = t3f_create_animation();
		if(ap)
		{
			t3f_animation_add_bitmap(ap, bp);
			t3f_animation_add_frame(ap, 0, 0, 0, 0, al_get_bitmap_width(bp), al_get_bitmap_height(bp), 0.0, 1, 0);
			t3f_animation_build_frame_list(ap);
		}
		else
		{
			al_destroy_bitmap(bp);
		}
	}
	return ap;
}

float vgolf_get_bitmap_width(T3F_ANIMATION * ap)
{
	return ap->frame[0]->width;
}

float vgolf_get_bitmap_height(T3F_ANIMATION * ap)
{
	return ap->frame[0]->height;
}

static char vgolf_date_string[1024] = {0};

char * vgolf_get_date_string(void)
{
	time_t etime = time(NULL);
	struct tm  * btime;
	btime = localtime(&etime);
	sprintf(vgolf_date_string, "%02d-%02d-%d - %02d%02d%02d", btime->tm_mon + 1, btime->tm_mday, 1900 + btime->tm_year, btime->tm_hour, btime->tm_min, btime->tm_sec);
	return vgolf_date_string;
}

static long vgolf_random_seed = 0;

void vgolf_srand(unsigned int seed)
{
	vgolf_random_seed = (long)seed;
}

#define VGOLF_RAND_MAX 0xFFFF
int vgolf_rand(void)
{
	return (((vgolf_random_seed = vgolf_random_seed * 214013L + 2531011L) >> 16) & VGOLF_RAND_MAX);
}

#define VGOLF_RS_SCALE (1.0 / (1.0 + VGOLF_RAND_MAX))
double vgolf_drand(void)
{
	double d;
	do
	{
		d = (((vgolf_rand() * VGOLF_RS_SCALE) + vgolf_rand()) * VGOLF_RS_SCALE + vgolf_rand()) * VGOLF_RS_SCALE;
	} while (d >= 1); /* round off */
	return d;
}

static int lc_cx, lc_cy, lc_cr, lc_collided;

void point_in_circle(ALLEGRO_BITMAP * bmp, int x, int y, int d)
{
	if(!lc_collided)
	{
		if(t3f_distance(lc_cx, lc_cy, x, y) < lc_cr)
		{
			lc_collided = 1;
		}
	}
}

void do_line(ALLEGRO_BITMAP *bmp, int x1, int y1, int x2, int y2, int d, void (*proc)(ALLEGRO_BITMAP *, int, int, int))
{
   int dx = x2-x1;
   int dy = y2-y1;
   int i1, i2;
   int x, y;
   int dd;

   /* worker macro */
   #define DO_LINE(pri_sign, pri_c, pri_cond, sec_sign, sec_c, sec_cond)     \
   {                                                                         \
      if (d##pri_c == 0) {                                                   \
	 proc(bmp, x1, y1, d);                                               \
	 return;                                                             \
      }                                                                      \
									     \
      i1 = 2 * d##sec_c;                                                     \
      dd = i1 - (sec_sign (pri_sign d##pri_c));                              \
      i2 = dd - (sec_sign (pri_sign d##pri_c));                              \
									     \
      x = x1;                                                                \
      y = y1;                                                                \
									     \
      while (pri_c pri_cond pri_c##2) {                                      \
	 proc(bmp, x, y, d);                                                 \
									     \
	 if (dd sec_cond 0) {                                                \
	    sec_c = sec_c sec_sign 1;                                        \
	    dd += i2;                                                        \
	 }                                                                   \
	 else                                                                \
	    dd += i1;                                                        \
									     \
	 pri_c = pri_c pri_sign 1;                                           \
      }                                                                      \
   }

   if (dx >= 0) {
      if (dy >= 0) {
	 if (dx >= dy) {
	    /* (x1 <= x2) && (y1 <= y2) && (dx >= dy) */
	    DO_LINE(+, x, <=, +, y, >=);
	 }
	 else {
	    /* (x1 <= x2) && (y1 <= y2) && (dx < dy) */
	    DO_LINE(+, y, <=, +, x, >=);
	 }
      }
      else {
	 if (dx >= -dy) {
	    /* (x1 <= x2) && (y1 > y2) && (dx >= dy) */
	    DO_LINE(+, x, <=, -, y, <=);
	 }
	 else {
	    /* (x1 <= x2) && (y1 > y2) && (dx < dy) */
	    DO_LINE(-, y, >=, +, x, >=);
	 }
      }
   }
   else {
      if (dy >= 0) {
	 if (-dx >= dy) {
	    /* (x1 > x2) && (y1 <= y2) && (dx >= dy) */
	    DO_LINE(-, x, >=, +, y, >=);
	 }
	 else {
	    /* (x1 > x2) && (y1 <= y2) && (dx < dy) */
	    DO_LINE(+, y, <=, -, x, <=);
	 }
      }
      else {
	 if (-dx >= -dy) {
	    /* (x1 > x2) && (y1 > y2) && (dx >= dy) */
	    DO_LINE(-, x, >=, -, y, <=);
	 }
	 else {
	    /* (x1 > x2) && (y1 > y2) && (dx < dy) */
	    DO_LINE(-, y, >=, -, x, <=);
	 }
      }
   }

   #undef DO_LINE
}

int line_circle_collide(float cx, float cy, float cr, float x1, float y1, float x2, float y2)
{
	float line_length;

	/* first see if the circle is close enough to the line for a possible collision */
	line_length = t3f_distance(x1, y1, x2, y2);
	if(t3f_distance(cx, cy, x1, y1) > line_length || t3f_distance(cx, cy, x2, y2) > line_length)
	{
		return 0;
	}

	lc_cx = cx;
	lc_cy = cy;
	lc_cr = cr;
	lc_collided = 0;
	do_line(NULL, x1, y1, x2, y2, 0, point_in_circle);
	return lc_collided;
}

void body_move(BODY * bo)
{
	if(!bo->moveable)
	{
		bo->vel.x = bo->vel.y = 0;
		return;
	}
	bo->pos.x += bo->vel.x;
	bo->pos.y += bo->vel.y;
}

//draw this polygon to the bitmap centered at this point
void draw_polygon(POLY po, int cx, int cy, float ox, float oy, float zoom, ALLEGRO_COLOR color)
{
	int i;

	for(i = 0; i < po.num_vertices - 1; i++)
	{
		al_draw_line((cx + po.p[i].x) * zoom - ox+ 0.5, (cy + po.p[i].y) * zoom - oy + 0.5, (cx + po.p[i+1].x) * zoom - ox + 0.5, (cy + po.p[i + 1].y) * zoom - oy + 0.5, color, 1.0 / t3f_display_scale_x);
	}
}

void draw_filled_polygon(POLY po, float cx, float cy, float ox, float oy, float zoom, ALLEGRO_COLOR color)
{
	int i;
	ALLEGRO_VERTEX v[64];

	for(i = 0; i < po.num_vertices; i++)
	{
		v[i].x = (cx + po.p[i].x) * zoom - ox;
		v[i].y = (cy + po.p[i].y) * zoom - oy;
		v[i].z = 0;
		v[i].color = color;
	}
	al_draw_prim(v, NULL, NULL, 0, po.num_vertices, ALLEGRO_PRIM_TRIANGLE_FAN);

}

void draw_poly2(ALLEGRO_BITMAP * bp, POLY po, VERTEX c, ALLEGRO_COLOR col)
{
	int i;
	float ang;
	VERTEX n, m;

	for(i = 0; i < po.num_vertices-1; i++)
	{
		ang = atan2(po.p[i].y-po.p[i+1].y, po.p[i].x-po.p[i+1].x)+M_PI/2;
		m = new_vertex((po.p[i].x+po.p[i+1].x)/2, (po.p[i].y+po.p[i+1].y)/2);
		n = new_vertex(10*cos(ang), 10*sin(ang));

		al_draw_line(c.x+po.p[i].x, c.y+po.p[i].y, c.x+po.p[i+1].x, c.y+po.p[i+1].y, col, 1.0);
		al_draw_line(c.x+m.x, c.y+m.y, c.x+m.x+n.x, c.y+m.y+n.y, col, 1.0);
	}
	al_draw_line(c.x+po.p[po.num_vertices-1].x, c.y+po.p[po.num_vertices-1].y, c.x+po.p[0].x, c.y+po.p[0].y, col, 1.0);
}

//tells if a collision has occured between these two bodies
char check_body_collision(BODY * a, BODY * b)
{
	return body_collide(a, b);
}

//checks to see if the ball is within a rectangle that surrounds the line segment endpoints..
char circle_lineseg_rect_collision(VERTEX b_c, float radius, VERTEX v1, VERTEX v2, VERTEX c)
{
	int x1,y1,x2,y2;

	//sort the points
	if(v1.x < v2.x)
	{
		x1 = v1.x;
		x2 = v2.x;
	}
	else
	{
		x2 = v1.x;
		x1 = v2.x;
	}
	if(v1.y < v2.y)
	{
		y1 = v1.y;
		y2 = v2.y;
	}
	else
	{
		y2 = v1.y;
		y1 = v2.y;
	}

	//make sure the rectangle is of a minimum span
	if(x2-x1 < MIN_RECT_SPAN)
	{
		x1 -= MIN_RECT_SPAN/2;
		x2 += MIN_RECT_SPAN/2;
	}
	if(y2-y1 < MIN_RECT_SPAN)
	{
		y1 -= MIN_RECT_SPAN/2;
		y2 += MIN_RECT_SPAN/2;
	}
	return rect_collision_f(-radius, -radius, radius, radius, x1, y1, x2, y2, b_c, c);
}

//tells if this poly is empty
char poly_is_empty(POLY po)
{
	return (po.color == POLY_TYPE_EMPTY_WALL);
}

//tells if this poly is empty
char poly_is_laser(POLY po)
{
	return (po.color == POLY_TYPE_LASER_WALL);
}

ALLEGRO_COLOR get_color(PALETTE * palette, int color)
{
	return al_map_rgb(palette->entry[color].r, palette->entry[color].g, palette->entry[color].b);
}

void triangle(float x1, float y1, float x2, float y2, float x3, float y3, ALLEGRO_COLOR color)
{
	ALLEGRO_VERTEX v[] =
	{
		{.x = x1, .y = y1, .z = 0, .color = color, .u = 0, .v = 0},
		{.x = x2, .y = y2, .z = 0, .color = color, .u = 0, .v = 0},
		{.x = x3, .y = y3, .z = 0, .color = color, .u = 0, .v = 0}
	};
	al_draw_prim(v, NULL, NULL, 0, 3, ALLEGRO_PRIM_TRIANGLE_LIST);
}

static bool compare_extension(const char * path, const char * ext)
{
	int i, start = 0;

	for(i = strlen(path) - 1; i >= 0; i--)
	{
		if(path[i] == '.')
		{
			start = i;
			break;
		}
	}
	for(i = 0; i < strlen(ext); i++)
	{
		if(path[start + i] != ext[i])
		{
			return false;
		}
	}
	return true;
}

int for_each_file_ex(const char *name, int in_attrib, int out_attrib, int (*callback)(const char *filename, int attrib, void *param), void *param)
{
	ALLEGRO_FS_ENTRY * dir;
	ALLEGRO_FS_ENTRY * file;
	int pos;
	int epos = 0;

	char extension[16] = {0};
	char path[1024] = {0};

	sprintf(path, "%s", name);
	pos = strlen(path);
	while(pos >= 0)
	{
		if(path[pos] == '*')
		{
			path[pos] = '\0';
			if(pos > 0)
			{
				path[pos - 1] = '\0';
			}
			break;
		}
		pos--;
	}
	for(pos = pos + 1; pos < strlen(name); pos++)
	{
		extension[epos] = path[pos];
		epos++;
	}
	extension[epos] = '\0';

	dir = al_create_fs_entry(path);
	if(dir)
	{
		if(al_open_directory(dir))
		{
			while(1)
			{
				file = al_read_directory(dir);
				if(file)
				{
					if(compare_extension(al_get_fs_entry_name(file), extension))
					{
						callback(al_get_fs_entry_name(file), 0, param);
					}
					al_destroy_fs_entry(file);
				}
				else
				{
					break;
				}
			}
		}
		al_destroy_fs_entry(dir);
	}
	return 1;
}
