#include "t3f/t3f.h"
#include "t3f/draw.h"
#include "t3f/menu.h"
#ifdef ALLEGRO_WINDOWS
	#include "t3f/windows.h"
#endif

#include "core/avatar.h"
#include "core/resource.h"

#include "main.h"
#include "ui_proc.h"
#include "undo.h"

bool app_setup_menus(APP_INSTANCE * app)
{
	app->file_menu = al_create_menu();
	if(!app->file_menu)
	{
		return false;
	}
	t3f_add_menu_item(app->file_menu, "&New", 0, NULL, ui_menu_proc_file_new, NULL);
	t3f_add_menu_item(app->file_menu, "&Open", 0, NULL, ui_menu_proc_file_open, NULL);
	t3f_add_menu_item(app->file_menu, "&Save", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_file_save, ui_update_menu);
	t3f_add_menu_item(app->file_menu, "Save &As", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_file_save_as, ui_update_menu);
	t3f_add_menu_item(app->file_menu, NULL, 0, NULL, NULL, NULL);
	t3f_add_menu_item(app->file_menu, "Edit &Tags", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_file_edit_tags, ui_update_menu);
	#ifndef ALLEGRO_MACOSX
		t3f_add_menu_item(app->file_menu, NULL, 0, NULL, NULL, NULL);
		t3f_add_menu_item(app->file_menu, NULL, 0, NULL, NULL, NULL);
		t3f_add_menu_item(app->file_menu, "&Quit", 0, NULL, ui_menu_proc_file_exit, NULL);
	#endif

	app->help_menu = al_create_menu();
	if(!app->help_menu)
	{
		return false;
	}
	t3f_add_menu_item(app->help_menu, "&Manual", 0, NULL, ui_menu_proc_help_manual, NULL);
	t3f_add_menu_item(app->help_menu, "&About", 0, NULL, ui_menu_proc_help_about, NULL);

	app->edit_menu = al_create_menu();
	if(!app->edit_menu)
	{
		return false;
	}
	t3f_add_menu_item(app->edit_menu, "&Undo", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_edit_undo, ui_update_menu_undo);
	t3f_add_menu_item(app->edit_menu, "&Redo", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_edit_redo, ui_update_menu_redo);
	t3f_add_menu_item(app->edit_menu, NULL, 0, NULL, NULL, NULL);
	t3f_add_menu_item(app->edit_menu, "C&ut", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_edit_cut, ui_update_menu);
	t3f_add_menu_item(app->edit_menu, "&Copy", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_edit_copy, ui_update_menu);
	t3f_add_menu_item(app->edit_menu, "&Paste", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_edit_paste, ui_update_menu_paste);

	app->tool_menu = al_create_menu();
	if(!app->tool_menu)
	{
		return false;
	}
	app->tool_menu_id[0] = t3f_add_menu_item(app->tool_menu, "Pencil", ALLEGRO_MENU_ITEM_CHECKED, NULL, ui_menu_proc_tool_pencil, ui_update_menu_tool);
	app->tool_menu_id[1] = t3f_add_menu_item(app->tool_menu, "Line", ALLEGRO_MENU_ITEM_CHECKBOX, NULL, ui_menu_proc_tool_line, ui_update_menu_tool);
	app->tool_menu_id[2] = t3f_add_menu_item(app->tool_menu, "Fill", ALLEGRO_MENU_ITEM_CHECKBOX, NULL, ui_menu_proc_tool_fill, ui_update_menu_tool);
	app->tool_menu_id[3] = t3f_add_menu_item(app->tool_menu, "Rectangle", ALLEGRO_MENU_ITEM_CHECKBOX, NULL, ui_menu_proc_tool_rectangle, ui_update_menu_tool);
	app->tool_menu_id[4] = t3f_add_menu_item(app->tool_menu, "Filled Rectangle", ALLEGRO_MENU_ITEM_CHECKBOX, NULL, ui_menu_proc_tool_filled_rectangle, ui_update_menu_tool);
	app->tool_menu_id[5] = t3f_add_menu_item(app->tool_menu, "Clear", ALLEGRO_MENU_ITEM_CHECKBOX, NULL, ui_menu_proc_tool_clear, ui_update_menu_tool);
	app->tool_menu_id[6] = t3f_add_menu_item(app->tool_menu, "Circle", ALLEGRO_MENU_ITEM_CHECKBOX, NULL, ui_menu_proc_tool_circle, ui_update_menu_tool);
	app->tool_menu_id[7] = t3f_add_menu_item(app->tool_menu, "Filled Circle", ALLEGRO_MENU_ITEM_CHECKBOX, NULL, ui_menu_proc_tool_filled_circle, ui_update_menu_tool);
	app->tool_menu_id[8] = t3f_add_menu_item(app->tool_menu, "Select", ALLEGRO_MENU_ITEM_CHECKBOX, NULL, ui_menu_proc_tool_select, ui_update_menu_tool);

	app->shift_menu = al_create_menu();
	if(!app->shift_menu)
	{
		return false;
	}
	t3f_add_menu_item(app->shift_menu, "Shift &Up", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_image_shift_up, ui_update_menu);
	t3f_add_menu_item(app->shift_menu, "Shift &Down", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_image_shift_down, ui_update_menu);
	t3f_add_menu_item(app->shift_menu, "Shift &Left", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_image_shift_left, ui_update_menu);
	t3f_add_menu_item(app->shift_menu, "Shift &Right", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_image_shift_right, ui_update_menu);

	app->turn_menu = al_create_menu();
	if(!app->turn_menu)
	{
		return false;
	}
	t3f_add_menu_item(app->turn_menu, "&Clocksise", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_image_turn_cw, ui_update_menu);
	t3f_add_menu_item(app->turn_menu, "C&ounter-Clocksise", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_image_turn_ccw, ui_update_menu);

	app->image_menu = al_create_menu();
	if(!app->image_menu)
	{
		return false;
	}
	t3f_add_menu_item(app->image_menu, "&Mirror", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_image_mirror, ui_update_menu);
	t3f_add_menu_item(app->image_menu, "&Flip", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_image_flip, ui_update_menu);
	t3f_add_menu_item(app->image_menu, NULL, 0, NULL, NULL, NULL);
	t3f_add_menu_item(app->image_menu, "Turn", ALLEGRO_MENU_ITEM_DISABLED, app->turn_menu, NULL, NULL);
	t3f_add_menu_item(app->image_menu, NULL, 0, NULL, NULL, NULL);
	t3f_add_menu_item(app->image_menu, "Shift", 0, app->shift_menu, NULL, NULL);
	t3f_add_menu_item(app->image_menu, NULL, 0, NULL, NULL, NULL);
	t3f_add_menu_item(app->image_menu, "Load &Image", ALLEGRO_MENU_ITEM_DISABLED, NULL, ui_menu_proc_image_load, ui_update_menu);

	app->menu = al_create_menu();
	if(!app->menu)
	{
		return false;
	}
	t3f_add_menu_item(app->menu, "&File", 0, app->file_menu, NULL, NULL);
	t3f_add_menu_item(app->menu, "&Edit", 0, app->edit_menu, NULL, NULL);
	t3f_add_menu_item(app->menu, "&Tool", 0, app->tool_menu, NULL, NULL);
	t3f_add_menu_item(app->menu, "&Image", 0, app->image_menu, NULL, NULL);
	t3f_add_menu_item(app->menu, "&Help", 0, app->help_menu, NULL, NULL);
	return true;
}

void app_destroy_menus(APP_INSTANCE * app)
{
	if(app->menu)
	{
		al_destroy_menu(app->menu);
		app->menu = NULL;
	}
	if(app->image_menu)
	{
		al_destroy_menu(app->image_menu);
		app->image_menu = NULL;
	}
	if(app->edit_menu)
	{
		al_destroy_menu(app->edit_menu);
		app->edit_menu = NULL;
	}
	if(app->file_menu)
	{
		al_destroy_menu(app->file_menu);
		app->file_menu = NULL;
	}
}

static bool app_setup_dialogs(APP_INSTANCE * app)
{
	int bezel = 8;
	int font_height;
	int text_width;
	int button_width;
	int box_height;
	int pos_y;

	app->box_theme = t3gui_load_theme("data/box_theme.ini", 0);
	if(!app->box_theme)
	{
		return false;
	}

	app->button_theme = t3gui_load_theme("data/button_theme.ini", 0);
	if(!app->button_theme)
	{
		return false;
	}

	app->list_box_theme = t3gui_load_theme("data/listbox_theme.ini", 0);
	if(!app->list_box_theme)
	{
		return false;
	}

	font_height = al_get_font_line_height(app->button_theme->state[0].font[0]);

	text_width = al_get_text_width(app->button_theme->state[0].font[0], "You have unsaved changes.");
	app->unsaved_dialog = t3gui_create_dialog();
	if(!app->unsaved_dialog)
	{
		return false;
	}
	button_width = text_width / 3;
	t3gui_dialog_add_element(app->unsaved_dialog, app->box_theme, t3gui_box_proc, 0, 0, bezel + text_width + bezel, bezel + font_height + bezel + 32 + bezel, 0, 0, 0, 0, NULL, NULL, NULL);
	t3gui_dialog_add_element(app->unsaved_dialog, app->button_theme, t3gui_text_proc, bezel, bezel, text_width, font_height, 0, 0, ALLEGRO_ALIGN_CENTRE, 0, "You have unsaved changes.", NULL, NULL);
	t3gui_dialog_add_element(app->unsaved_dialog, app->button_theme, t3gui_push_button_proc, bezel, bezel + font_height + bezel, button_width, 32, 0, D_EXIT, 0, 0, "Save", ui_proc_unsaved_save, NULL);
	t3gui_dialog_add_element(app->unsaved_dialog, app->button_theme, t3gui_push_button_proc, bezel + button_width, bezel + font_height + bezel, button_width, 32, 0, D_EXIT, 0, 0, "Discard", ui_proc_unsaved_discard, NULL);
	t3gui_dialog_add_element(app->unsaved_dialog, app->button_theme, t3gui_push_button_proc, bezel + button_width * 2, bezel + font_height + bezel, button_width, 32, 0, D_EXIT, 0, 0, "Cancel", NULL, NULL);
	t3gui_center_dialog(app->unsaved_dialog, al_get_display_width(t3f_display), al_get_display_height(t3f_display));

	text_width = al_get_text_width(app->button_theme->state[0].font[0], app->copyright_message);
	app->about_dialog = t3gui_create_dialog();
	if(!app->about_dialog)
	{
		return false;
	}
	t3gui_dialog_add_element(app->about_dialog, app->box_theme, t3gui_box_proc, 0, 0, bezel + text_width + bezel, bezel + font_height * 2 + bezel + 32 + bezel, 0, 0, 0, 0, NULL, NULL, NULL);
	t3gui_dialog_add_element(app->about_dialog, app->button_theme, t3gui_text_proc, bezel, bezel, text_width, font_height, 0, 0, ALLEGRO_ALIGN_CENTRE, 0, T3F_APP_TITLE " v" T3F_APP_VERSION, NULL, NULL);
	t3gui_dialog_add_element(app->about_dialog, app->button_theme, t3gui_text_proc, bezel, bezel + font_height, text_width, font_height, 0, 0, ALLEGRO_ALIGN_CENTRE, 0, app->copyright_message, NULL, NULL);
	t3gui_dialog_add_element(app->about_dialog, app->button_theme, t3gui_push_button_proc, bezel + text_width / 2 - 32, bezel + font_height * 2 + bezel, 64, 32, 0, D_EXIT, 0, 0, "Okay", NULL, NULL);
	t3gui_center_dialog(app->about_dialog, al_get_display_width(t3f_display), al_get_display_height(t3f_display));

	app->tags_dialog = t3gui_create_dialog();
	if(!app->tags_dialog)
	{
		return false;
	}
	box_height = bezel + font_height + bezel + font_height + 8 + bezel + font_height + bezel + font_height + 8 + bezel + font_height + bezel + font_height + 8 + bezel + 32 + bezel;
	t3gui_dialog_add_element(app->tags_dialog, app->box_theme, t3gui_box_proc, 0, 0, 320, box_height, 0, 0, 0, 0, NULL, NULL, NULL);
	pos_y = bezel;
	t3gui_dialog_add_element(app->tags_dialog, app->button_theme, t3gui_text_proc, bezel, pos_y, 320 - bezel * 2, font_height, 0, 0, 0, 0, "Name", NULL, NULL);
	pos_y += font_height + bezel;
	t3gui_dialog_add_element(app->tags_dialog, app->list_box_theme, t3gui_edit_proc, bezel, pos_y, 320 - bezel * 2, font_height, 0, 0, 256, 0, app->tags_edit_text[0], NULL, NULL);
	pos_y += font_height + 8 + bezel;
	t3gui_dialog_add_element(app->tags_dialog, app->button_theme, t3gui_text_proc, bezel, pos_y, 320 - bezel * 2, font_height, 0, 0, 0, 0, "Author", NULL, NULL);
	pos_y += font_height + bezel;
	t3gui_dialog_add_element(app->tags_dialog, app->list_box_theme, t3gui_edit_proc, bezel, pos_y, 320 - bezel * 2, font_height, 0, 0, 256, 0, app->tags_edit_text[1], NULL, NULL);
	pos_y += font_height + 8 + bezel;
	t3gui_dialog_add_element(app->tags_dialog, app->button_theme, t3gui_text_proc, bezel, pos_y, 320 - bezel * 2, font_height, 0, 0, 0, 0, "Comment", NULL, NULL);
	pos_y += font_height + bezel;
	t3gui_dialog_add_element(app->tags_dialog, app->list_box_theme, t3gui_edit_proc, bezel, pos_y, 320 - bezel * 2, font_height, 0, 0, 1024, 0, app->tags_edit_text[2], NULL, NULL);
	pos_y += font_height + 8 + bezel;
	t3gui_dialog_add_element(app->tags_dialog, app->button_theme, t3gui_push_button_proc, bezel, box_height - bezel - 32, 160 - bezel - bezel / 2, 32, 0, D_EXIT, 0, 0, "Okay", ui_proc_tags_okay, NULL);
	t3gui_dialog_add_element(app->tags_dialog, app->button_theme, t3gui_push_button_proc, 160 + bezel / 2, box_height - bezel - 32, 160 - bezel - bezel / 2, 32, 0, D_EXIT, 0, 0, "Cancel", ui_proc_tags_cancel, NULL);
	t3gui_center_dialog(app->tags_dialog, al_get_display_width(t3f_display), al_get_display_height(t3f_display));

	return true;
}

static void app_destroy_dialogs(APP_INSTANCE * app)
{
	t3gui_destroy_dialog(app->about_dialog);
	t3gui_destroy_dialog(app->unsaved_dialog);
}


static bool enlarge_window(APP_INSTANCE * app)
{
	ALLEGRO_DISPLAY * new_display;
	ALLEGRO_MONITOR_INFO info;
	bool recreate = false;
	int desktop_width;
	int desktop_height;
	int window_multiplier_x;
	int window_multiplier_y;
	int window_multiplier;
	int window_width;
	int window_height;

	al_get_monitor_info(0, &info);
	desktop_width = info.x2 - info.x1;
	desktop_height = info.y2 - info.y1;
	window_multiplier_x = desktop_width / 408;
	if(window_multiplier_x < 1)
	{
		window_multiplier_x = 1;
	}
	window_multiplier_y = desktop_height / 424;
	if(window_multiplier_y < 1)
	{
		window_multiplier_y = 1;
	}
	if(window_multiplier_x < window_multiplier_y)
	{
		window_multiplier = window_multiplier_x;
	}
	else
	{
		window_multiplier = window_multiplier_y;
	}
	if(window_multiplier < 2)
	{
		return true;
	}
	window_width = 408 * window_multiplier;
	window_height = 424 * window_multiplier;
	#ifdef ALLEGRO_UNIX
		#ifndef ALLEGRO_MACOSX
			recreate = true;
		#endif
	#endif
	if(recreate)
	{
		new_display = al_create_display(window_width, window_height);
		if(!new_display)
		{
			return false;
		}
		al_destroy_display(t3f_display);
		t3f_display = new_display;
		al_register_event_source(t3f_queue, al_get_display_event_source(t3f_display));
		t3f_get_base_transform();
		t3f_default_view->need_update = true;
		t3f_select_view(t3f_default_view);
	}
	else
	{
		t3f_set_gfx_mode(window_width, window_height, t3f_flags);
	}
	return true;
}

/* initialize our app, load graphics, etc. */
bool app_initialize(APP_INSTANCE * app, int argc, char * argv[])
{
	int i;

	memset(app, 0, sizeof(APP_INSTANCE));
	/* initialize T3F */
	if(!t3f_initialize(T3F_APP_TITLE, 408, 424, 60.0, app_logic, app_render, T3F_DEFAULT | T3F_USE_MENU, app))
	{
		printf("Error initializing T3F\n");
		return false;
	}
	if(!enlarge_window(app))
	{
		return false;
	}

	#ifdef ALLEGRO_WINDOWS
		t3f_windows_text_to_utf8(T3F_APP_COPYRIGHT, app->copyright_message, 256);
	#else
		strcpy(app->copyright_message, T3F_APP_COPYRIGHT);
	#endif
	ui_load_config(app);
	al_set_new_bitmap_flags(ALLEGRO_NO_PREMULTIPLIED_ALPHA);
	al_set_blender(ALLEGRO_ADD, ALLEGRO_ALPHA, ALLEGRO_INVERSE_ALPHA);

	for(i = 1; i < argc; i++)
	{
		if(!strcmp(argv[i], "--make_preview"))
		{
			if(i + 1 >= argc)
			{
				printf("Usage: vgd-avatar-editor --make_preview <avatar_file>\n");
				return false;
			}
			else
			{
				ALLEGRO_PATH * pp;

				app->avatar = vgolf_load_avatar(argv[i + 1], NULL, NULL);
				if(!app->avatar)
				{
					printf("Unable to load avatar %s\n", argv[i + 1]);
					return false;
				}
				pp = al_create_path(argv[i + 1]);
				if(pp)
				{
					al_set_path_extension(pp, ".t3a");
					t3f_save_animation(app->avatar->image[0], al_path_cstr(pp, '/'));
					vgolf_destroy_avatar(app->avatar);
					al_destroy_path(pp);
				}
				return false;
			}
		}
	}

	if(!t3gui_init())
	{
		return false;
	}
	if(!app_setup_menus(app))
	{
		return false;
	}
	t3f_attach_menu(app->menu);

	if(!app_setup_dialogs(app))
	{
		return false;
	}

	vgolf_register_resource_loaders();

	/* load data */
	app->font = t3f_generate_font("data/Signpgsc.ttf", 32, T3F_FONT_OUTLINE);
	if(!app->font)
	{
		return false;
	}
	app->font->scale = 0.5;
	app->tool_bitmap[0] = al_load_bitmap("data/drawing_tool_0.pcx");
	if(!app->tool_bitmap[0])
	{
		return false;
	}
	app->tool_bitmap[1] = al_load_bitmap("data/drawing_tool_1.pcx");
	if(!app->tool_bitmap[1])
	{
		return false;
	}
	app->tool_bitmap[2] = al_load_bitmap("data/drawing_tool_2.pcx");
	if(!app->tool_bitmap[2])
	{
		return false;
	}
	app->tool_bitmap[3] = al_load_bitmap("data/drawing_tool_3.pcx");
	if(!app->tool_bitmap[3])
	{
		return false;
	}
	app->tool_bitmap[4] = al_load_bitmap("data/drawing_tool_4.pcx");
	if(!app->tool_bitmap[4])
	{
		return false;
	}
	app->tool_bitmap[5] = al_load_bitmap("data/drawing_tool_5.pcx");
	if(!app->tool_bitmap[5])
	{
		return false;
	}
	app->tool_bitmap[6] = al_load_bitmap("data/drawing_tool_6.pcx");
	if(!app->tool_bitmap[6])
	{
		return false;
	}
	app->tool_bitmap[7] = al_load_bitmap("data/drawing_tool_7.pcx");
	if(!app->tool_bitmap[7])
	{
		return false;
	}
	app->tool_bitmap[8] = al_load_bitmap("data/drawing_tool_8.pcx");
	if(!app->tool_bitmap[8])
	{
		return false;
	}
	app->palette_tool_bitmap = al_load_bitmap("data/drawing_palette_gui.pcx");
	if(!app->palette_tool_bitmap)
	{
		return false;
	}
	app->palette_bitmap = al_load_bitmap("data/drawing_palette.pcx");
	if(!app->palette_bitmap)
	{
		return false;
	}
	app->bg_bitmap = al_load_bitmap("data/drawing_bg.pcx");
	if(!app->bg_bitmap)
	{
		return false;
	}
	app->scratch_bitmap = al_create_bitmap(64, 58);
	if(!app->scratch_bitmap)
	{
		return false;
	}
	app->editor_bitmap = al_create_bitmap(64, 58);
	if(!app->editor_bitmap)
	{
		return false;
	}
	ui_menu_proc_file_new(0, app);
	t3f_set_event_handler(app_event_handler);
	return true;
}
