#include "t3f/t3f.h"
#include "t3f/sound.h"
#include "t3f/menu.h"
#include "main.h"
#include "undo.h"
#include "ui_proc.h"
#include "core/commentary.h"

int ui_update_menu(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->avatar)
    {
        al_set_menu_item_flags(menu, item, 0);
    }
    else
    {
        al_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    return 0;
}

int ui_update_menu_undo(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->avatar && undo_available(data))
    {
        al_set_menu_item_flags(menu, item, 0);
    }
    else
    {
        al_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    return 0;
}

int ui_update_menu_redo(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->avatar && redo_available(data))
    {
        al_set_menu_item_flags(menu, item, 0);
    }
    else
    {
        al_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    return 0;
}

int ui_update_menu_tool(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    int i;
    int selected_item = -1;

    for(i = 0; i < 9; i++)
    {
        if(app->tool_menu_id[i] == item)
        {
            selected_item = i;
            break;
        }
    }
    if(selected_item == app->selected_tool)
    {
        al_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_CHECKED);
    }
    else
    {
        al_set_menu_item_flags(menu, item, 0);
    }
    return 0;
}

int ui_update_menu_paste(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->avatar && app->clipboard_bitmap)
    {
        al_set_menu_item_flags(menu, item, 0);
    }
    else
    {
        al_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    return 0;
}

static void ui_save_config(void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->last_open_path)
    {
        al_set_config_value(t3f_config, "UI", "last_open_path", al_path_cstr(app->last_open_path, '/'));
    }
    if(app->last_open_image_path)
    {
        al_set_config_value(t3f_config, "UI", "last_open_image_path", al_path_cstr(app->last_open_image_path, '/'));
    }
}

void ui_load_config(void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    const char * val;

    if(app->last_open_path)
    {
        al_destroy_path(app->last_open_path);
    }
    if(app->last_open_image_path)
    {
        al_destroy_path(app->last_open_image_path);
    }
    val = al_get_config_value(t3f_config, "UI", "last_open_path");
    if(val)
    {
        app->last_open_path = al_create_path(val);
    }
    val = al_get_config_value(t3f_config, "UI", "last_open_image_path");
    if(val)
    {
        app->last_open_image_path = al_create_path(val);
    }
}

static int ui_menu_proc_file_new_helper(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_STATE old_state;
    int i;

    app_unselect(data);
    if(app->avatar)
    {
        vgolf_destroy_avatar(app->avatar);
    }
    app->avatar = malloc(sizeof(VGOLF_AVATAR));
    if(app->avatar)
    {
        memset(app->avatar, 0, sizeof(VGOLF_AVATAR));
        for(i = 0; i < 4; i++)
        {
            app->avatar->image[i] = t3f_create_animation();
            if(app->avatar->image[i])
            {
                app->avatar->image[i]->bitmaps->bitmap[0] = al_create_bitmap(64, 58);
                if(app->avatar->image[i]->bitmaps->bitmap[0])
                {
                    al_store_state(&old_state, ALLEGRO_STATE_TARGET_BITMAP);
                    al_set_target_bitmap(app->avatar->image[i]->bitmaps->bitmap[0]);
                    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
                    al_restore_state(&old_state);
                    app->avatar->image[i]->bitmaps->count++;
                    t3f_animation_add_frame(app->avatar->image[i], 0, 0.0, 0.0, 0.0, -1.0, -1.0, 0.0, 1, 0);
                }
                app->float_bp = app->avatar->image[0]->bitmaps->bitmap[0];
            }
        }
    }
    undo_reset(data);
    redo_reset(data);
    if(app->last_open_path)
    {
        al_destroy_path(app->last_open_path);
        app->last_open_path = NULL;
    }
    app_update_window_title(data);
    t3f_refresh_menus();
    return 0;
}

int ui_menu_proc_file_new(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->changes)
    {
        app->unsaved_dialog_trigger = 1;
        t3gui_show_dialog(app->unsaved_dialog, NULL, 0, app);
    }
    else
    {
        ui_menu_proc_file_new_helper(id, data);
    }
    return 0;
}

static int ui_menu_proc_file_open_helper(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_FILECHOOSER * fp;
    const char * initial = NULL;

    if(app->last_open_path)
    {
        initial = al_path_cstr(app->last_open_path, '/');
    }
    fp = al_create_native_file_dialog(initial, "Open Avatar File", "*.vda;*.pcx;*.png", ALLEGRO_FILECHOOSER_FILE_MUST_EXIST);
    if(fp)
    {
        if(al_show_native_file_dialog(t3f_display, fp))
        {
            if(al_get_native_file_dialog_count(fp))
            {
                if(app->last_open_path)
                {
                    al_destroy_path(app->last_open_path);
                }
                app->last_open_path = al_create_path(al_get_native_file_dialog_path(fp, 0));
                app_unselect(data);
                if(app->avatar)
                {
                    vgolf_destroy_avatar(app->avatar);
                }
                app->avatar = vgolf_load_avatar(al_path_cstr(app->last_open_path, '/'), NULL, NULL);
                undo_reset(data);
                redo_reset(data);
                app_update_window_title(data);
                t3f_refresh_menus();
                ui_save_config(data);
            }
        }
    }
    return 0;
}

int ui_menu_proc_file_open(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->changes)
    {
        app->unsaved_dialog_trigger = 2;
        t3gui_show_dialog(app->unsaved_dialog, NULL, 0, app);
    }
    else
    {
        ui_menu_proc_file_open_helper(id, data);
    }
    return 0;
}

int ui_menu_proc_file_save(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_PATH * preview_path;

    app_unselect(data);
    if(!app->last_open_path)
    {
        ui_menu_proc_file_save_as(id, data);
    }
    if(app->last_open_path)
    {
        vgolf_save_avatar(app->avatar, NULL, al_path_cstr(app->last_open_path, '/'));
        preview_path = al_clone_path(app->last_open_path);
        if(preview_path)
        {
            al_set_path_extension(preview_path, ".t3a");
            t3f_save_animation(app->avatar->image[0], al_path_cstr(preview_path, '/'));
            al_destroy_path(preview_path);
        }
        app->changes = 0;
        app_update_window_title(data);
        return 1;
    }
    return 0;
}

int ui_menu_proc_file_save_as(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_FILECHOOSER * fp;
    const char * initial = NULL;

    if(app->last_open_path)
    {
        initial = al_path_cstr(app->last_open_path, '/');
    }

    fp = al_create_native_file_dialog(initial, "Save Avatar As", "*.png", ALLEGRO_FILECHOOSER_SAVE);
    if(fp)
    {
        if(al_show_native_file_dialog(t3f_display, fp))
        {
            if(al_get_native_file_dialog_count(fp))
            {
                if(app->last_open_path)
                {
                    al_destroy_path(app->last_open_path);
                }
                app->last_open_path = al_create_path(al_get_native_file_dialog_path(fp, 0));
                al_set_path_extension(app->last_open_path, ".png");
                ui_menu_proc_file_save(id, data);
                ui_save_config(data);
            }
        }
    }
    return 0;
}

int ui_menu_proc_file_edit_tags(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    strcpy(app->tags_edit_text[0], app->avatar->name);
    strcpy(app->tags_edit_text[1], app->avatar->author);
    strcpy(app->tags_edit_text[2], app->avatar->comment);
    t3gui_show_dialog(app->tags_dialog, NULL, 0, data);
    return 0;
}

int ui_menu_proc_file_exit(int id, void * data)
{
    ui_proc_try_exit(id, data);
    return 0;
}

int ui_menu_proc_edit_undo(int id, void * data)
{
    app_unselect(data);
    undo_restore(data);
    t3f_refresh_menus();
    return 0;
}

int ui_menu_proc_edit_redo(int id, void * data)
{
    app_unselect(data);
    redo_restore(data);
    t3f_refresh_menus();
    return 0;
}

int ui_menu_proc_edit_cut(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_STATE old_state;
    ALLEGRO_TRANSFORM identity;

    if(app->clipboard_bitmap)
    {
        al_destroy_bitmap(app->clipboard_bitmap);
        app->clipboard_bitmap = NULL;
    }
    if(app->selection_active)
    {
        app->clipboard_bitmap = al_create_bitmap(app->selection_right - app->selection_left, app->selection_bottom - app->selection_top);
        if(app->clipboard_bitmap)
        {
            al_store_state(&old_state, ALLEGRO_STATE_TRANSFORM | ALLEGRO_STATE_TARGET_BITMAP | ALLEGRO_STATE_BLENDER);
            al_set_target_bitmap(app->clipboard_bitmap);
            al_identity_transform(&identity);
            al_use_transform(&identity);
            al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
            al_draw_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0], -app->selection_left, -app->selection_top, 0);
            al_set_target_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]);
            al_use_transform(&identity);
            al_set_blender(ALLEGRO_ADD, ALLEGRO_ONE, ALLEGRO_ZERO);
            al_draw_filled_rectangle(app->selection_left, app->selection_top, app->selection_right, app->selection_bottom, al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
            al_restore_state(&old_state);
            t3f_refresh_menus();
        }
    }
    else
    {
        app->clipboard_bitmap = al_clone_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]);
        if(app->clipboard_bitmap)
        {
            al_store_state(&old_state, ALLEGRO_STATE_TRANSFORM | ALLEGRO_STATE_TARGET_BITMAP | ALLEGRO_STATE_BLENDER);
            al_set_target_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]);
            al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
            al_restore_state(&old_state);
            t3f_refresh_menus();
        }
    }
    return 0;
}

int ui_menu_proc_edit_copy(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_STATE old_state;
    ALLEGRO_TRANSFORM identity;

    if(app->clipboard_bitmap)
    {
        al_destroy_bitmap(app->clipboard_bitmap);
        app->clipboard_bitmap = NULL;
    }
    if(app->selection_active)
    {
        app->clipboard_bitmap = al_create_bitmap(app->selection_right - app->selection_left, app->selection_bottom - app->selection_top);
        if(app->clipboard_bitmap)
        {
            al_store_state(&old_state, ALLEGRO_STATE_TRANSFORM | ALLEGRO_STATE_TARGET_BITMAP);
            al_set_target_bitmap(app->clipboard_bitmap);
            al_identity_transform(&identity);
            al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
            al_draw_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0], -app->selection_left, -app->selection_top, 0);
            al_restore_state(&old_state);
            t3f_refresh_menus();
        }
    }
    else
    {
        app->clipboard_bitmap = al_clone_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]);
        t3f_refresh_menus();
    }
    return 0;
}

int ui_menu_proc_edit_paste(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_TRANSFORM identity;
    ALLEGRO_STATE old_state;

    app_unselect(data);
    undo_store(app);
    app->float_left = 0;
    app->float_top = 0;
    app->float_right = al_get_bitmap_width(app->clipboard_bitmap);
    app->float_bottom = al_get_bitmap_height(app->clipboard_bitmap);
    app->selection_active = true;
    app->selection_left = -1;
    app->selection_right = -1;
    app->selection_top = -1;
    app->selection_bottom = -1;
    app->selection_active = true;
    app->float_original = al_clone_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]);
    app->float_bitmap = al_clone_bitmap(app->clipboard_bitmap);
    app->float_bp = app->avatar->image[app->selected_image]->bitmaps->bitmap[0];
    al_store_state(&old_state, ALLEGRO_STATE_TARGET_BITMAP);
    al_set_target_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]);
    al_identity_transform(&identity);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_bitmap(app->clipboard_bitmap, 0, 0, 0);
    al_restore_state(&old_state);
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_tool_pencil(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    app->selected_tool = 0;
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_tool_line(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    app->selected_tool = 1;
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_tool_fill(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    app->selected_tool = 2;
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_tool_rectangle(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    app->selected_tool = 3;
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_tool_filled_rectangle(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    app->selected_tool = 4;
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_tool_clear(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    app->selected_tool = 5;
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_tool_circle(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    app->selected_tool = 6;
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_tool_filled_circle(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    app->selected_tool = 7;
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_tool_select(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    app->selected_tool = 8;
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_image_mirror(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_TRANSFORM identity;
    ALLEGRO_STATE old_state;

    undo_store(app);
    al_store_state(&old_state, ALLEGRO_STATE_TARGET_BITMAP);
    al_set_target_bitmap(app->scratch_bitmap);
    al_identity_transform(&identity);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0], 0, 0, ALLEGRO_FLIP_HORIZONTAL);
    al_set_target_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_bitmap(app->scratch_bitmap, 0, 0, 0);
    al_restore_state(&old_state);
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_image_flip(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_TRANSFORM identity;
    ALLEGRO_STATE old_state;

    undo_store(app);
    al_store_state(&old_state, ALLEGRO_STATE_TARGET_BITMAP);
    al_set_target_bitmap(app->scratch_bitmap);
    al_identity_transform(&identity);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0], 0, 0, ALLEGRO_FLIP_VERTICAL);
    al_set_target_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_bitmap(app->scratch_bitmap, 0, 0, 0);
    al_restore_state(&old_state);
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_image_turn_cw(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_TRANSFORM identity;
    ALLEGRO_STATE old_state;

    undo_store(app);
    al_store_state(&old_state, ALLEGRO_STATE_TARGET_BITMAP);
    al_set_target_bitmap(app->scratch_bitmap);
    al_identity_transform(&identity);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_rotated_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0], al_get_bitmap_width(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]) / 2, al_get_bitmap_height(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]) / 2, al_get_bitmap_width(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]) / 2, al_get_bitmap_height(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]) / 2, ALLEGRO_PI / 2.0, 0);
    al_set_target_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_bitmap(app->scratch_bitmap, 0, 0, 0);
    al_restore_state(&old_state);
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_image_turn_ccw(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_TRANSFORM identity;
    ALLEGRO_STATE old_state;

    undo_store(app);
    al_store_state(&old_state, ALLEGRO_STATE_TARGET_BITMAP);
    al_set_target_bitmap(app->scratch_bitmap);
    al_identity_transform(&identity);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_rotated_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0], al_get_bitmap_width(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]) / 2, al_get_bitmap_height(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]) / 2, al_get_bitmap_width(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]) / 2, al_get_bitmap_height(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]) / 2, -ALLEGRO_PI / 2.0, 0);
    al_set_target_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_bitmap(app->scratch_bitmap, 0, 0, 0);
    al_restore_state(&old_state);
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_image_shift_up(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_TRANSFORM identity;
    ALLEGRO_STATE old_state;

    undo_store(app);
    al_store_state(&old_state, ALLEGRO_STATE_TARGET_BITMAP);
    al_set_target_bitmap(app->scratch_bitmap);
    al_identity_transform(&identity);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0], 0, -1, 0);
    al_draw_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0], 0, al_get_bitmap_height(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]) - 1, 0);
    al_set_target_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_bitmap(app->scratch_bitmap, 0, 0, 0);
    al_restore_state(&old_state);
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_image_shift_down(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_TRANSFORM identity;
    ALLEGRO_STATE old_state;

    undo_store(app);
    al_store_state(&old_state, ALLEGRO_STATE_TARGET_BITMAP);
    al_set_target_bitmap(app->scratch_bitmap);
    al_identity_transform(&identity);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0], 0, 1, 0);
    al_draw_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0], 0, -al_get_bitmap_height(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]) + 1, 0);
    al_set_target_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_bitmap(app->scratch_bitmap, 0, 0, 0);
    al_restore_state(&old_state);
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_image_shift_left(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_TRANSFORM identity;
    ALLEGRO_STATE old_state;

    undo_store(app);
    al_store_state(&old_state, ALLEGRO_STATE_TARGET_BITMAP);
    al_set_target_bitmap(app->scratch_bitmap);
    al_identity_transform(&identity);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0], -1, 0, 0);
    al_draw_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0], al_get_bitmap_width(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]) - 1, 0, 0);
    al_set_target_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_bitmap(app->scratch_bitmap, 0, 0, 0);
    al_restore_state(&old_state);
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_image_shift_right(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_TRANSFORM identity;
    ALLEGRO_STATE old_state;

    undo_store(app);
    al_store_state(&old_state, ALLEGRO_STATE_TARGET_BITMAP);
    al_set_target_bitmap(app->scratch_bitmap);
    al_identity_transform(&identity);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0], 1, 0, 0);
    al_draw_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0], -al_get_bitmap_width(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]) + 1, 0, 0);
    al_set_target_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]);
    al_use_transform(&identity);
    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
    al_draw_bitmap(app->scratch_bitmap, 0, 0, 0);
    al_restore_state(&old_state);
    t3f_refresh_menus();

    return 0;
}

int ui_menu_proc_image_load(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_BITMAP * bp;
    ALLEGRO_TRANSFORM identity;
    ALLEGRO_STATE old_state;
    ALLEGRO_FILECHOOSER * fp;
    const char * initial = NULL;

    if(app->last_open_image_path)
    {
        initial = al_path_cstr(app->last_open_image_path, '/');
    }

    fp = al_create_native_file_dialog(initial, "Load Image", "*.png;*.pcx;*.bmp", ALLEGRO_FILECHOOSER_FILE_MUST_EXIST);
    if(fp)
    {
        if(al_show_native_file_dialog(t3f_display, fp))
        {
            if(al_get_native_file_dialog_count(fp))
            {
                if(app->last_open_image_path)
                {
                    al_destroy_path(app->last_open_image_path);
                }
                app->last_open_image_path = al_create_path(al_get_native_file_dialog_path(fp, 0));
                ui_save_config(data);
                bp = al_load_bitmap(al_path_cstr(app->last_open_image_path, '/'));
                if(bp)
                {
                    undo_store(app);
                    al_store_state(&old_state, ALLEGRO_STATE_TARGET_BITMAP | ALLEGRO_STATE_TRANSFORM);
                    al_set_target_bitmap(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]);
                    al_identity_transform(&identity);
                    al_use_transform(&identity);
                    al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
                    al_draw_scaled_bitmap(bp, 0, 0, al_get_bitmap_width(bp), al_get_bitmap_height(bp), 0, 0, al_get_bitmap_width(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]), al_get_bitmap_height(app->avatar->image[app->selected_image]->bitmaps->bitmap[0]), 0);
                    al_restore_state(&old_state);
                    t3f_refresh_menus();
                }
            }
        }
    }
    return 0;
}

int ui_menu_proc_help_manual(int id, void * data)
{
    int ret;

    #ifdef ALLEGRO_WINDOWS
		ret = system("start data\\manual\\index.htm");
	#else
		#ifdef ALLEGRO_MACOSX
			ret = system("open data/manual/index.htm");
		#else
			ret = system("xdg-open data/manual/index.htm");
		#endif
	#endif
    return ret;
}

int ui_menu_proc_help_about(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    t3gui_show_dialog(app->about_dialog, NULL, 0, NULL);
    return 0;
}

int ui_proc_try_exit(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->changes)
    {
        app->unsaved_dialog_trigger = 0;
        t3gui_show_dialog(app->unsaved_dialog, NULL, 0, app);
    }
    else
    {
        t3f_exit();
    }
    return 0;
}

static void finish_unsaved(APP_INSTANCE * app)
{
    switch(app->unsaved_dialog_trigger)
    {
        case 1:
        {
            ui_menu_proc_file_new_helper(0, app);
            break;
        }
        case 2:
        {
            ui_menu_proc_file_open_helper(0, app);
            break;
        }
        default:
        {
            t3f_exit();
            break;
        }
    }
}

int ui_proc_unsaved_save(T3GUI_ELEMENT * d, void * dp)
{
    APP_INSTANCE * app = (APP_INSTANCE *)dp;

    if(ui_menu_proc_file_save(0, dp))
    {
        finish_unsaved(app);
    }

    return 1;
}

int ui_proc_unsaved_discard(T3GUI_ELEMENT * d, void * dp)
{
    APP_INSTANCE * app = (APP_INSTANCE *)dp;

    finish_unsaved(app);

    return 1;
}

int ui_proc_unsaved_cancel(T3GUI_ELEMENT * d, void * dp)
{
    return 1;
}

int ui_proc_tags_okay(T3GUI_ELEMENT * d, void * dp)
{
    APP_INSTANCE * app = (APP_INSTANCE *)dp;

    strcpy(app->avatar->name, app->tags_edit_text[0]);
    strcpy(app->avatar->author, app->tags_edit_text[1]);
    strcpy(app->avatar->comment, app->tags_edit_text[2]);

    return 1;
}

int ui_proc_tags_cancel(T3GUI_ELEMENT * d, void * dp)
{
    return 1;
}
