#include "t3f/resource.h"
#include "course.h"
#include "legacy_course.h"
#include "file.h"
#include "packfile.h"
#include "pcx.h"
#include "atlas.h"
#include "utility.h"
#include "resource.h"

/*static ALLEGRO_BITMAP * load_pcx_cond(void ** bp, ALLEGRO_FILE * fp, const char * fn)
{
	if(al_fgetc(fp))
	{
		return t3f_load_resource_f(bp, VGOLF_RESOURCE_TYPE_PCX, fp, "", 0, 0);
	}
	printf("failed to load image\n");
	return NULL;
} */

static T3F_ANIMATION * load_pcx_ani_cond(ALLEGRO_FILE * fp, const char * fn)
{
	T3F_ANIMATION * ap;
	if(al_fgetc(fp))
	{
		ap = t3f_create_animation();
		if(ap)
		{
			ap->bitmaps->bitmap[0] = NULL;
			t3f_load_resource_f((void **)&ap->bitmaps->bitmap[0], VGOLF_RESOURCE_TYPE_PCX, fp, fn, 0, 0);
			if(ap->bitmaps->bitmap[0])
			{
				ap->bitmaps->count = 1;
				t3f_animation_add_frame(ap, 0, 0, 0, 0, al_get_bitmap_width(ap->bitmaps->bitmap[0]), al_get_bitmap_height(ap->bitmaps->bitmap[0]), 0, 1, 0);
				ap->frame[0]->flags = 0;
				return ap;
			}
		}
	}
	return NULL;
}

static void skip_pcx_cond(ALLEGRO_FILE * fp)
{
	ALLEGRO_BITMAP * bp;

	if(al_fgetc(fp))
	{
		bp = load_pcx_f(fp);
		if(bp)
		{
			al_destroy_bitmap(bp);
		}
	}
}

int vgolf_load_legacy_hole_fp(HOLE * hole, ALLEGRO_FILE * in_file, const char * fn, PALETTE * pp)
{
	int i, j, k;
	char temp_char[4];
	int version;

	//read header..
	al_fread(in_file, temp_char, 4);
	if(temp_char[0] != 'C') return 0;
	if(temp_char[1] != 'R') return 0;
	if(temp_char[2] != 'S') return 0;
	if(temp_char[3] > 36) return 0;
	version = temp_char[3];

	hole->s.x = 320;
	hole->s.y = 240;
	hole->deccel_factor = 0.05f;

	switch(version)
	{
		case 36:
		{
			for(i = 0; i < 128; i++) hole->name[i] = al_fgetc(in_file);
			hole->par_score = al_fread16le(in_file);
			hole->body_layers = 1;
			hole->bodies[0] = al_fread16le(in_file);
			if(hole->bodies[0] > MAX_HOLE_POLYS)
			{
				printf("poly out of bounds\n");
				return 0;
			}

			for(i = 0; i < hole->bodies[0]; i++)
			{
				hole->body[0][i].poly.num_vertices = al_fread16le(in_file);
				if(hole->body[0][i].poly.num_vertices > MAX_OBJECT_VERTICES)
				{
					printf("vertex out of bounds\n");
					return 0;
				}
				hole->body[0][i].poly.color = al_fread16le(in_file);
				for(j = 0; j < hole->body[0][i].poly.num_vertices; j++)
				{
					hole->body[0][i].poly.p[j].x = al_fread16le(in_file);
					if(al_fgetc(in_file)) hole->body[0][i].poly.p[j].x *= -1;
					hole->body[0][i].poly.p[j].y = al_fread16le(in_file);
					if(al_fgetc(in_file)) hole->body[0][i].poly.p[j].y *= -1;
					al_fread16le(in_file);
					al_fgetc(in_file);
//					if(j > 0)
//					{ //do some correction so we don't get vertical lines..
//						if(hole->body[0][i].poly.p[j-1].x == hole->body[0][i].poly.p[j].x)
//						{
//							hole->body[0][i].poly.p[j].x += 0.5;
//						}
//					}
				}
				hole->body[0][i].friction = 0;
				hole->body[0][i].restitution = hole->body[0][i].mass = hole->body[0][i].inv_mass = 0;
				hole->body[0][i].density = 1;
				hole->body[0][i].pos = hole->s;
				hole->body[0][i].vel.x = hole->body[0][i].vel.y = 0;
				hole->body[0][i].glue = 0;
			}

			hole->num_verts = al_fread16le(in_file);
			for(i = 0; i < hole->num_verts; i++)
			{
				hole->vert[i].s.x = al_fread16le(in_file);
				if(al_fgetc(in_file)) hole->vert[i].s.x *= -1;
				hole->vert[i].s.y = al_fread16le(in_file);
				if(al_fgetc(in_file)) hole->vert[i].s.y *= -1;
				al_fread16le(in_file);
				al_fgetc(in_file);
				hole->vert[i].flag = al_fread16le(in_file);
				hole->vert[i].radius = al_fread16le(in_file);
				hole->vert[i].num_refs = al_fread16le(in_file);
				for(j = 0; j < hole->vert[i].num_refs; j++)
				{
					hole->vert[i].ref_no[j] = al_fread16le(in_file);
				}
				hole->vert[i].layer = 0;
				hole->vert[i].tick = 0;
			}

			for(i = 0; i < MAX_HOLE_LAYERS; i++)
			{
				hole->layer_pic_offset[0][i] = al_ftell(in_file);
				hole->layer_pic_offset[1][i] = al_ftell(in_file);
				hole->layer_pic[i] = t3f_load_resource_f((void **)&hole->layer_pic[i], VGOLF_RESOURCE_TYPE_RAW, in_file, fn, i, 0);
				t3f_destroy_resource(hole->layer_pic[i]);
				hole->layer_pic[i] = NULL;
			}
			break;
		}
		case 35: // deluxe intermediate format
		{
			for(i = 0; i < 128; i++) hole->name[i] = al_fgetc(in_file);
			hole->par_score = al_fread16le(in_file); // 0
			hole->body_layers = al_fgetc(in_file); // 2
			for(k = 0; k < hole->body_layers; k++)
			{
				hole->bodies[k] = al_fread16le(in_file); // 3
				if(hole->bodies[k] > MAX_HOLE_POLYS)
				{
					printf("poly out of bounds\n");
					return 0;
				}
				for(i = 0; i < hole->bodies[k]; i++)
				{
					hole->body[k][i].poly.num_vertices = al_fread16le(in_file); // 5
					if(hole->body[k][i].poly.num_vertices > MAX_OBJECT_VERTICES)
					{
						printf("vertex out of bounds\n");
						return 0;
					}
					hole->body[k][i].poly.color = al_fread16le(in_file); // 7
					for(j = 0; j < hole->body[k][i].poly.num_vertices; j++)
					{
						hole->body[k][i].poly.p[j].x = al_fread16le(in_file); // 9
						if(al_fgetc(in_file)) hole->body[k][i].poly.p[j].x *= -1; // 11
						hole->body[k][i].poly.p[j].y = al_fread16le(in_file); // 12
						if(al_fgetc(in_file)) hole->body[k][i].poly.p[j].y *= -1; // 14
//						if(j > 0)
//						{ //do some correction so we don't get vertical lines..
//							if(hole->body[k][i].poly.p[j-1].x == hole->body[k][i].poly.p[j].x)
//							{
//								hole->body[k][i].poly.p[j].x += 0.5;
//							}
//						}
					}
					hole->body[k][i].friction = 0;
					hole->body[k][i].restitution = hole->body[k][i].mass = hole->body[k][i].inv_mass = 0;
					hole->body[k][i].density = 1;
					hole->body[k][i].pos = hole->s;
					hole->body[k][i].vel.x = hole->body[k][i].vel.y = 0;
					hole->body[k][i].glue = 0;
				}
			}

			hole->num_verts = al_fread16le(in_file);
			for(i = 0; i < hole->num_verts; i++)
			{
				hole->vert[i].s.x = al_fread16le(in_file);
				if(al_fgetc(in_file)) hole->vert[i].s.x *= -1;
				hole->vert[i].s.y = al_fread16le(in_file);
				if(al_fgetc(in_file)) hole->vert[i].s.y *= -1;
				hole->vert[i].flag = al_fread16le(in_file);
				hole->vert[i].radius = al_fread16le(in_file);
				hole->vert[i].num_refs = al_fread16le(in_file);
				for(j = 0; j < hole->vert[i].num_refs; j++)
				{
					hole->vert[i].ref_no[j] = al_fread16le(in_file);
				}
				hole->vert[i].layer = al_fgetc(in_file);
				hole->vert[i].tick = 0;
			}

			for(i = 0; i < MAX_HOLE_LAYERS; i++)
			{
				hole->layer_pic_offset[0][i] = al_ftell(in_file);
				hole->layer_pic_offset[1][i] = al_ftell(in_file);
				hole->layer_pic[i] = t3f_load_resource_f((void **)&hole->layer_pic[i], VGOLF_RESOURCE_TYPE_RAW, in_file, fn, i, 0);
				t3f_destroy_resource(hole->layer_pic[i]);
				hole->layer_pic[i] = NULL;
			}
			break;
		}
	}
	return 1;
}

bool vgolf_is_legacy_course(const char * fn)
{
	ALLEGRO_FILE * fp;
	bool legacy = false;
	char header[16] = {0};

	fp = vgolf_fopen_packfile(fn, "rp");
	if(fp)
	{
		if(al_fread(fp, header, 16) == 16)
		{
			if(!strcasecmp(header, "vGolf_crs"))
			{
				legacy = true;
			}
		}
		al_fclose(fp);
	}
	return legacy;
}

void vgolf_read_legacy_course_info(VGOLF_COURSE * cr, ALLEGRO_FILE * fp, const char * fn)
{
	/* course info */
	al_fread(fp, cr->name, 128);
	al_fread(fp, cr->author, 256);
	al_fread(fp, cr->comment, 1024);
	cr->num_holes = al_fread32le(fp);
	cr->course_par = al_fread32le(fp);
	cr->image_bank.iconpic = load_pcx_ani_cond(fp, fn);
}

VGOLF_COURSE * vgolf_load_legacy_course_info(const char * filename, const ALLEGRO_FILE_INTERFACE * fi)
{
	VGOLF_COURSE * cr = NULL;
	ALLEGRO_FILE * fp = NULL;
	char header[16] = {0};
	const ALLEGRO_FILE_INTERFACE * old_fi;

	if(fi)
	{
		old_fi = al_get_new_file_interface();
		al_set_new_file_interface(fi);
	}

	fp = vgolf_fopen_packfile(filename, "rp");
	if(!fp)
	{
		goto cleanup;
	}
	if(al_fread(fp, header, 16) != 16)
	{
		printf("failed to read header\n");
	}
	if(strcasecmp(header, "vGolf_crs"))
	{
		goto cleanup;
	}

	cr = malloc(sizeof(VGOLF_COURSE));
	if(!cr)
	{
		goto cleanup;
	}
	memset(cr, 0, sizeof(VGOLF_COURSE));
	vgolf_read_legacy_course_info(cr, fp, filename);

	cleanup:
	{
		if(fp)
		{
			al_fclose(fp);
		}
		if(fi)
		{
			al_set_new_file_interface(old_fi);
		}
	}
	return cr;
}

VGOLF_COURSE * vgolf_load_legacy_course(const char * filename, int type, int flags, const ALLEGRO_FILE_INTERFACE * fi)
{
	VGOLF_COURSE * cr = NULL;
	ALLEGRO_FILE * fp = NULL;
	ALLEGRO_FILE * gfp = NULL;
	char header[16] = {0};
	int i, j;
	const ALLEGRO_FILE_INTERFACE * old_fi;

	if(fi)
	{
		old_fi = al_get_new_file_interface();
		al_set_new_file_interface(fi);
	}

	fp = vgolf_fopen_packfile(filename, "rp");
	if(!fp)
	{
		goto cleanup;
	}
	al_fread(fp, header, 16);
	if(strcasecmp(header, "vGolf_crs"))
	{
		goto cleanup;
	}

	cr = malloc(sizeof(VGOLF_COURSE));
	if(!cr)
	{
		goto cleanup;
	}
	memset(cr, 0, sizeof(VGOLF_COURSE));

	if(flags & VGOLF_COURSE_LOAD_EDIT)
	{
		al_set_new_bitmap_flags(ALLEGRO_NO_PREMULTIPLIED_ALPHA | ALLEGRO_MIN_LINEAR | ALLEGRO_MAG_LINEAR);
	}

	vgolf_read_legacy_course_info(cr, fp, filename);
	/* course data */
	//tee
	cr->image_bank.vert_pic[0][0] = load_pcx_ani_cond(fp, filename);
	//hole
	cr->image_bank.vert_pic[1][0] = load_pcx_ani_cond(fp, filename);
	cr->image_bank.vert_pic[1][1] = load_pcx_ani_cond(fp, filename);
	//trees
	cr->image_bank.object_pic[0][0] = load_pcx_ani_cond(fp, filename);
	cr->image_bank.object_pic[1][0] = load_pcx_ani_cond(fp, filename);
	//laser/wall
	cr->image_bank.object_pic[2][0] = load_pcx_ani_cond(fp, filename);
	cr->image_bank.object_pic[2][1] = load_pcx_ani_cond(fp, filename);
	//teleport pad
	cr->image_bank.object_pic[3][0] = load_pcx_ani_cond(fp, filename);
	cr->image_bank.object_pic[3][1] = load_pcx_ani_cond(fp, filename);
	//conveyor belt
	cr->image_bank.object_pic[4][0] = load_pcx_ani_cond(fp, filename);
	cr->image_bank.object_pic[4][1] = load_pcx_ani_cond(fp, filename);
	cr->image_bank.object_pic[4][2] = load_pcx_ani_cond(fp, filename);
	cr->image_bank.object_pic[4][3] = load_pcx_ani_cond(fp, filename);
	//animals
	cr->image_bank.animal_pic[0][0] = load_pcx_ani_cond(fp, filename);
	cr->image_bank.animal_pic[0][1] = load_pcx_ani_cond(fp, filename);
	cr->image_bank.animal_pic[0][2] = load_pcx_ani_cond(fp, filename);
	cr->image_bank.animal_pic[0][3] = load_pcx_ani_cond(fp, filename);
	cr->image_bank.animal_pic[1][0] = load_pcx_ani_cond(fp, filename);
	cr->image_bank.animal_pic[1][1] = load_pcx_ani_cond(fp, filename);
	cr->image_bank.animal_pic[1][2] = load_pcx_ani_cond(fp, filename);
	cr->image_bank.animal_pic[1][3] = load_pcx_ani_cond(fp, filename);

	//load victory screen images
	cr->victory_graphics_offset[0] = al_ftell(fp);
	cr->victory_graphics_offset[1] = al_ftell(fp);
	skip_pcx_cond(fp);
	skip_pcx_cond(fp);
	skip_pcx_cond(fp);
	skip_pcx_cond(fp);
	skip_pcx_cond(fp);
	skip_pcx_cond(fp);
	skip_pcx_cond(fp);
	//load scoreboard images
	cr->scoreboard.bgpic = load_pcx_ani_cond(fp, filename);
	memcpy(&cr->palette, &current_palette, sizeof(PALETTE));
	vgolf_set_resource_palette(&cr->palette);
	for(i = 0; i < cr->num_holes; i++)
	{
		vgolf_load_legacy_hole_fp(&cr->hole[i], fp, filename, &cr->palette);
	}

	/* correct tree radii */
	for(i = 0; i < cr->num_holes; i++)
	{
		for(j = 0; j < cr->hole[i].num_verts; j++)
		{
			if(cr->hole[i].vert[j].flag == VERT_FLAG_TREE_0 && cr->hole[i].vert[j].radius <= 20)
			{
				cr->hole[i].vert[j].radius = vgolf_get_bitmap_width(cr->image_bank.object_pic[0][0]) / 2;
			}
			else if(cr->hole[i].vert[j].flag == VERT_FLAG_TREE_1 && cr->hole[i].vert[j].radius <= 20)
			{
				cr->hole[i].vert[j].radius = vgolf_get_bitmap_width(cr->image_bank.object_pic[1][0]) / 2;
			}
		}
	}
	//width and height of the numbers on the scoreboard
	cr->scoreboard.num_w = 17;
	cr->scoreboard.num_h = 17;
	//x,y values of the scoreboard for positioning on the screen
	cr->scoreboard.x = 0;
	cr->scoreboard.y = 0;
	//x,y offset values of the course name on the scoreboard
	cr->scoreboard.cname_x = 20;
	cr->scoreboard.cname_y = 15;
	//x,y offset values of the menu items on the scoreboard
	cr->scoreboard.menu_x[0] = 130;
	cr->scoreboard.menu_y[0] = 440;
	cr->scoreboard.menu_x[1] = 420;
	cr->scoreboard.menu_y[1] = 440;
	//x,y offset values of the grids on the scoreboard (player 1-4)
	cr->scoreboard.grid_x[0] = 20 + 30;
	cr->scoreboard.grid_y[0] = 50 + 8;
	cr->scoreboard.grid_x[1] = 20 + 320 + 10;
	cr->scoreboard.grid_y[1] = 50 + 8;
	cr->scoreboard.grid_x[2] = 20 + 30;
	cr->scoreboard.grid_y[2] = 250;
	cr->scoreboard.grid_x[3] = 20 + 320 + 10;
	cr->scoreboard.grid_y[3] = 250;
	//# of segments on scoreboard and # of active ones
	cr->scoreboard.num_scoreboard_segs = 10;
	cr->scoreboard.scoreboard_seg_count = 0;
	cr->scoreboard.scoreboard_seg_counter = 0;
	cr->scoreboard.scoreboard_seg_counter_end = 48;

	cr->legacy = true;

	/* optionally load victory graphics */
	if(flags & VGOLF_COURSE_LOAD_ALL)
	{
		vgolf_load_legacy_course_victory_graphics(cr, filename, type, flags, fi);
	}

	/* optionally load all hole graphics */
	if(flags & VGOLF_COURSE_LOAD_ALL)
	{
		for(i = 0; i < cr->num_holes; i++)
		{
			for(j = 0; j < MAX_HOLE_LAYERS; j++)
			{
				gfp = vgolf_fopen_packfile(filename, "rp");
				if(gfp)
				{
					al_fseek(gfp, cr->hole[i].layer_pic_offset[type][j], ALLEGRO_SEEK_CUR);
					t3f_load_resource_f((void **)&cr->hole[i].layer_pic[j], VGOLF_RESOURCE_TYPE_RAW, gfp, filename, j, 0);
					al_fclose(gfp);
				}
			}
		}
	}

	cleanup:
	{
		if(fp)
		{
			al_fclose(fp);
		}
		if(fi)
		{
			al_set_new_file_interface(old_fi);
		}
	}

	return cr;
}

int vgolf_load_legacy_hole_graphics(HOLE * hole, const char * fn, int type, const ALLEGRO_FILE_INTERFACE * fi)
{
	ALLEGRO_FILE * gfp = NULL;
	int j;
	int ret = 0;
	const ALLEGRO_FILE_INTERFACE * old_fi;

	if(fi)
	{
		old_fi = al_get_new_file_interface();
		al_set_new_file_interface(fi);
	}

	for(j = 0; j < MAX_HOLE_LAYERS; j++)
	{
		gfp = vgolf_fopen_packfile(fn, "rp");
		if(gfp)
		{
			al_fseek(gfp, hole->layer_pic_offset[type][j], ALLEGRO_SEEK_CUR);
			hole->layer_pic[j] = t3f_load_resource_f((void **)&hole->layer_pic[j], VGOLF_RESOURCE_TYPE_RAW, gfp, fn, j, 0);
		}
		else
		{
			goto cleanup;
		}
		ret = 1;
	}

	cleanup:
	{
		if(gfp)
		{
			al_fclose(gfp);
		}
		if(fi)
		{
			al_set_new_file_interface(old_fi);
		}
	}
	return ret;
}

int vgolf_load_legacy_course_victory_graphics(VGOLF_COURSE * cr, const char * filename, int type, int flags, const ALLEGRO_FILE_INTERFACE * fi)
{
	ALLEGRO_FILE * fp = NULL;
	int ret = 0;
	const ALLEGRO_FILE_INTERFACE * old_fi;

	if(fi)
	{
		old_fi = al_get_new_file_interface();
		al_set_new_file_interface(fi);
	}

	fp = vgolf_fopen_packfile(filename, "rp");
	if(fp)
	{
		al_fseek(fp, cr->victory_graphics_offset[type], ALLEGRO_SEEK_CUR);
		cr->image_bank.victory_bgpic = load_pcx_ani_cond(fp, filename);
		cr->image_bank.victory_podiumpic = load_pcx_ani_cond(fp, filename);
		cr->image_bank.victory_cloudpic[0] = load_pcx_ani_cond(fp, filename);
		cr->image_bank.victory_cloudpic[1] = load_pcx_ani_cond(fp, filename);
		cr->image_bank.victory_trophypic[0] = load_pcx_ani_cond(fp, filename);
		cr->image_bank.victory_trophypic[1] = load_pcx_ani_cond(fp, filename);
		cr->image_bank.victory_trophypic[2] = load_pcx_ani_cond(fp, filename);
		al_fclose(fp);
		ret = 1;
	}

	if(fi)
	{
		al_set_new_file_interface(old_fi);
	}
	return ret;
}
