#include "t3f/t3f.h"
#include "t3f/sound.h"
#include "t3f/menu.h"
#include "main.h"
#include "core/commentary.h"

static char ui_last_open_path[1024] = {0};
static char ui_last_open_audio_path[1024] = {0};

int ui_menu_update_item(ALLEGRO_MENU * menu, int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->commentary)
    {
        al_set_menu_item_flags(menu, id, 0);
    }
    else
    {
        al_set_menu_item_flags(menu, id, ALLEGRO_MENU_ITEM_DISABLED);
    }
    return 0;
}

static void ui_save_config(void)
{
    al_set_config_value(t3f_config, "UI", "last_open_path", ui_last_open_path);
    al_set_config_value(t3f_config, "UI", "last_open_audio_path", ui_last_open_audio_path);
}

void ui_load_config(void)
{
    const char * val;

    val = al_get_config_value(t3f_config, "UI", "last_open_path");
    if(val)
    {
        strncpy(ui_last_open_path, val, 1024);
    }
    val = al_get_config_value(t3f_config, "UI", "last_open_audio_path");
    if(val)
    {
        strncpy(ui_last_open_audio_path, val, 1024);
    }
}

void ui_update_buttons(APP_INSTANCE * app)
{
    int i;

    for(i = 0; app->dialog[0]->element[i].proc; i++)
    {
        if(app->dialog[0]->element[i].proc == t3gui_push_button_proc)
        {
            if(app->commentary)
            {
                if(!strcmp(app->dialog[0]->element[i].dp, "Add"))
                {
                    if(app->commentary->comments[app->dialog[0]->element[1].d1].count < MAX_MESSAGE_STRINGS)
                    {
                        app->dialog[0]->element[i].flags = 0;
                    }
                    else
                    {
                        app->dialog[0]->element[i].flags = D_DISABLED;
                    }
                }
                else if(!strcmp(app->dialog[0]->element[i].dp, "Delete"))
                {
                    if(app->commentary->comments[app->dialog[0]->element[1].d1].count)
                    {
                        app->dialog[0]->element[i].flags = 0;
                    }
                    else
                    {
                        app->dialog[0]->element[i].flags = D_DISABLED;
                    }
                }
                else if(!strcmp(app->dialog[0]->element[i].dp, "Export"))
                {
                    if(app->commentary->comments[app->dialog[0]->element[1].d1].count)
                    {
                        app->dialog[0]->element[i].flags = 0;
                    }
                    else
                    {
                        app->dialog[0]->element[i].flags = D_DISABLED;
                    }
                }
                else if(!strcmp(app->dialog[0]->element[i].dp, "Listen"))
                {
                    if(app->commentary->comments[app->dialog[0]->element[1].d1].count)
                    {
                        app->dialog[0]->element[i].flags = 0;
                    }
                    else
                    {
                        app->dialog[0]->element[i].flags = D_DISABLED;
                    }
                }
            }
            else
            {
                app->dialog[0]->element[i].flags = 0;
            }
        }
    }
}

int ui_menu_proc_file_new(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->commentary)
    {
        destroy_vgolf_commentary(app->commentary);
    }
    app->commentary = malloc(sizeof(VGOLF_COMMENTARY));
    if(app->commentary)
    {
        memset(app->commentary, 0, sizeof(VGOLF_COMMENTARY));
    }
    t3f_refresh_menus();
    ui_update_buttons(app);
    return 0;
}

int ui_menu_proc_file_open(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_FILECHOOSER * fp;

    fp = al_create_native_file_dialog(ui_last_open_path, "Open Commentary File", "*.vc", ALLEGRO_FILECHOOSER_FILE_MUST_EXIST);
    if(fp)
    {
        if(al_show_native_file_dialog(t3f_display, fp))
        {
            if(al_get_native_file_dialog_count(fp))
            {
                strncpy(ui_last_open_path, al_get_native_file_dialog_path(fp, 0), 1024);
                if(app->commentary)
                {
                    destroy_vgolf_commentary(app->commentary);
                }
                app->commentary = load_vgolf_commentary(ui_last_open_path, NULL);
                ui_save_config();
                t3f_refresh_menus();
                ui_update_buttons(app);
            }
        }
    }
    return 0;
}

int ui_menu_proc_file_save(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    save_vgolf_commentary(app->commentary, ui_last_open_path);
    return 0;
}

int ui_menu_proc_file_save_as(int id, void * data)
{
    ALLEGRO_FILECHOOSER * fp;

    fp = al_create_native_file_dialog(ui_last_open_path, "Save Commentary As", "*.vc", ALLEGRO_FILECHOOSER_SAVE);
    if(fp)
    {
        if(al_show_native_file_dialog(t3f_display, fp))
        {
            if(al_get_native_file_dialog_count(fp))
            {
                strncpy(ui_last_open_path, al_get_native_file_dialog_path(fp, 0), 1024);
                ui_menu_proc_file_save(id, data);
                ui_save_config();
            }
        }
    }
    return 0;
}

int ui_menu_proc_file_tags(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    strncpy(app->tags_title, app->commentary->name, 256);
    strncpy(app->tags_author, app->commentary->author, 256);
    strncpy(app->tags_comment, app->commentary->comment, 1024);
    t3gui_show_dialog(app->dialog[1], t3f_queue, 0, app);
    app->state = STATE_EDIT_TAGS;
    return 0;
}

int ui_menu_proc_file_exit(int id, void * data)
{
    t3f_exit();
    return 0;
}

int ui_menu_proc_help_manual(int id, void * data)
{
    return 0;
}

int ui_menu_proc_help_about(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    t3gui_show_dialog(app->about_dialog, t3f_queue, 0, app);
    return 0;
}

char * ui_type_list_proc(int index, int * list_size, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    if(index < 0)
    {
        if(list_size)
        {
            *list_size = app->commentary ? MAX_MESSAGE_STRING_TYPES : 0;
        }
        return NULL;
    }
    if(app->commentary)
    {
    	switch(index)
    	{
    		case 0:
    		{
    			return "Finish Good";
    		}
    		case 1:
    		{
    			return "Finish Better";
    		}
    		case 2:
    		{
    			return "Finish Great";
    		}
    		case 3:
    		{
    			return "Finish Excellent";
    		}
    		case 4:
    		{
    			return "Finish Poor";
    		}
    		case 5:
    		{
    			return "Finish Very Poor";
    		}
    		case 6:
    		{
    			return "Ball Zapped";
    		}
    		case 7:
    		{
    			return "Close Shot";
    		}
    		case 8:
    		{
    			return "Water Hazard";
    		}
    		case 9:
    		{
    			return "Ball Crushed";
    		}
    	}
    }
	return NULL;
}

static char ui_voice_text[256] = {0};

char * ui_voice_list_proc(int index, int *list_size, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    if(index < 0)
    {
        if(list_size)
        {
            *list_size = app->commentary ? app->commentary->comments[app->dialog[0]->element[1].d1].count : 0;
        }
        return NULL;
    }
    if(app->commentary)
    {
       sprintf(ui_voice_text, "Voice %d", index + 1);
       return ui_voice_text;
   }
   return NULL;
}

int ui_voice_add_button_proc(T3GUI_ELEMENT * d, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_FILECHOOSER * fp;

    fp = al_create_native_file_dialog(ui_last_open_audio_path, "Open Audio File", "*.wav;*.ogg", ALLEGRO_FILECHOOSER_FILE_MUST_EXIST);
    if(fp)
    {
        if(al_show_native_file_dialog(t3f_display, fp))
        {
            if(al_get_native_file_dialog_count(fp))
            {
                strncpy(ui_last_open_audio_path, al_get_native_file_dialog_path(fp, 0), 1024);
                app->commentary->comments[app->dialog[0]->element[1].d1].comment[app->commentary->comments[app->dialog[0]->element[1].d1].count].voice = al_load_sample(ui_last_open_audio_path);
                if(app->commentary->comments[app->dialog[0]->element[1].d1].comment[app->commentary->comments[app->dialog[0]->element[1].d1].count].voice)
                {
                    app->commentary->comments[app->dialog[0]->element[1].d1].count++;
                }
                ui_save_config();
                ui_update_buttons(app);
            }
        }
    }
    return 1;
}

int ui_voice_delete_button_proc(T3GUI_ELEMENT * d, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    int i;

    if(app->dialog[0]->element[2].d1 < app->commentary->comments[app->dialog[0]->element[1].d1].count)
    {
        for(i = app->dialog[0]->element[2].d1; i < app->commentary->comments[app->dialog[0]->element[1].d1].count - 1; i++)
        {
            memcpy(&app->commentary->comments[app->dialog[0]->element[1].d1].comment[i], &app->commentary->comments[app->dialog[0]->element[1].d1].comment[i + 1], sizeof(VGOLF_COMMENT));
        }
        app->commentary->comments[app->dialog[0]->element[1].d1].count--;
        ui_update_buttons(app);
    }
    return 1;
}

int ui_voice_export_button_proc(T3GUI_ELEMENT * d, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_FILECHOOSER * fp;

    fp = al_create_native_file_dialog(ui_last_open_audio_path, "Save Audio File", "*.wav", ALLEGRO_FILECHOOSER_SAVE);
    if(fp)
    {
        if(al_show_native_file_dialog(t3f_display, fp))
        {
            if(al_get_native_file_dialog_count(fp))
            {
                strncpy(ui_last_open_audio_path, al_get_native_file_dialog_path(fp, 0), 1024);
                al_save_sample(ui_last_open_audio_path, app->commentary->comments[app->dialog[0]->element[1].d1].comment[app->dialog[0]->element[2].d1].voice);
                ui_save_config();
            }
        }
    }
    return 1;
}

int ui_voice_listen_button_proc(T3GUI_ELEMENT * d, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    t3f_play_sample(app->commentary->comments[app->dialog[0]->element[1].d1].comment[app->dialog[0]->element[2].d1].voice, 1.0, 0.0, 1.0);
    return 1;
}
