#include "t3f/t3f.h"
#include "t3f/resource.h"
#include "pcx.h"
#include "resource.h"
#include "packfile.h"

static PALETTE * vgolf_resource_palette = NULL;

static void * vgolf_pcx_resource_handler_proc(ALLEGRO_FILE * fp, const char * filename, int option, int flags, unsigned long offset)
{
	void * ptr = NULL;
	ALLEGRO_STATE old_state;
	bool openfp = false; // operating on already open file

	al_store_state(&old_state, ALLEGRO_STATE_NEW_BITMAP_PARAMETERS);
	al_set_new_bitmap_flags(al_get_new_bitmap_flags() | ALLEGRO_NO_PRESERVE_TEXTURE);
	if(fp)
	{
		openfp = true;
	}
	if(!openfp && offset == 0)
	{
		ptr = load_pcx(filename);
	}
	else
	{
		if(!openfp)
		{
			fp = vgolf_fopen_packfile(filename, "rp");
			al_fseek(fp, offset, ALLEGRO_SEEK_SET);
		}
		if(fp)
		{
			ptr = load_pcx_f(fp);
			if(!openfp)
			{
				al_fclose(fp);
			}
		}
	}
	al_restore_state(&old_state);
	return ptr;
}

static void vgolf_pcx_resource_handler_destroy_proc(void * ptr)
{
	al_destroy_bitmap(ptr);
}

static ALLEGRO_BITMAP * vgolf_load_raw_image_f(ALLEGRO_FILE * fp, int option)
{
	ALLEGRO_LOCKED_REGION *lr;
	int temp_w, temp_h;
	int c, j, k;
	ALLEGRO_BITMAP * ptr = NULL;

	temp_w = al_fread16le(fp);
	temp_h = al_fread16le(fp);
	if(temp_w != 0 && temp_h != 0)
	{
		ptr = al_create_bitmap(temp_w, temp_h);
		if(!ptr)
		{
			return NULL;
		}
		lr = al_lock_bitmap(ptr, ALLEGRO_PIXEL_FORMAT_ABGR_8888_LE, ALLEGRO_LOCK_WRITEONLY);
		if(lr)
		{
			for(j = 0; j < temp_h; j++)
			{
				unsigned char *dest = (unsigned char*)lr->data + j*lr->pitch;
				for(k = 0; k < temp_w; k++)
				{
					c = al_fgetc(fp);
					dest[k*4    ] = vgolf_resource_palette->entry[c].r;
					dest[k*4 + 1] = vgolf_resource_palette->entry[c].g;
					dest[k*4 + 2] = vgolf_resource_palette->entry[c].b;
					dest[k*4 + 3] = option > 0 ? (c == 0 ? 0 : 255) : 255;
				}
			}
			al_unlock_bitmap(ptr);
		}
	}
	return ptr;
}

static ALLEGRO_BITMAP * vgolf_load_raw_image(const char * fn, int option)
{
	ALLEGRO_FILE * fp;
	ALLEGRO_BITMAP * bp = NULL;

	fp = al_fopen(fn, "rb");
	if(fp)
	{
		bp = vgolf_load_raw_image_f(fp, option);
		al_fclose(fp);
	}
	return bp;
}

static void * vgolf_raw_resource_handler_proc(ALLEGRO_FILE * fp, const char * filename, int option, int flags, unsigned long offset)
{
	void * ptr = NULL;
	ALLEGRO_STATE old_state;
	bool openfp = false; // operating on already open file

	al_store_state(&old_state, ALLEGRO_STATE_NEW_BITMAP_PARAMETERS);
	al_set_new_bitmap_flags(al_get_new_bitmap_flags() | ALLEGRO_NO_PRESERVE_TEXTURE);
	if(fp)
	{
		openfp = true;
	}
	if(!openfp && offset == 0)
	{
		ptr = vgolf_load_raw_image(filename, option);
	}
	else
	{
		if(!openfp)
		{
			fp = vgolf_fopen_packfile(filename, "rp");
			al_fseek(fp, offset, ALLEGRO_SEEK_SET);
		}
		if(fp)
		{
			ptr = vgolf_load_raw_image_f(fp, option);
			if(!openfp)
			{
				al_fclose(fp);
			}
		}
	}
	al_restore_state(&old_state);
	return ptr;
}

static void vgolf_raw_resource_handler_destroy_proc(void * ptr)
{
	al_destroy_bitmap(ptr);
}

void vgolf_set_resource_palette(PALETTE * pp)
{
	vgolf_resource_palette = pp;
}

void vgolf_register_resource_loaders(void)
{
	t3f_register_resource_handler(VGOLF_RESOURCE_TYPE_PCX, vgolf_pcx_resource_handler_proc, vgolf_pcx_resource_handler_destroy_proc);
	t3f_register_resource_handler(VGOLF_RESOURCE_TYPE_RAW, vgolf_raw_resource_handler_proc, vgolf_raw_resource_handler_destroy_proc);
}
