#include "t3f/t3f.h"

#include "core/poly.h"
#include "instance.h"

void crsed_set_clipping_rectangle(ALLEGRO_TRANSFORM * tp, int x, int y, int w, int h)
{
	float tx, ty;
	float twx, twy;
	float ox = 0.0, oy = 0.0;

	/* convert virtual screen coordinates to real display coordinates */
	al_transform_coordinates(tp, &ox, &oy);
	if(w != 0 && h != 0)
	{
		tx = x;
		ty = y;
		twx = w;
		twy = h;
	}
	else
	{
		tx = 0;
		ty = 0;
		twx = t3f_virtual_display_width;
		twy = t3f_virtual_display_height - 16;
	}
	al_transform_coordinates(tp, &tx, &ty);
	al_transform_coordinates(tp, &twx, &twy);
	al_set_clipping_rectangle(tx, ty, twx - ox, twy - oy);
}

void crsed_setup_transformation(void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	float r, vr;
	float w, h;
	float tw, th;

	tw = al_get_display_width(t3f_display);
	th = al_get_display_height(t3f_display) - 16;
	r = th / tw;
	vr = (float)(t3f_virtual_display_height - 16.0) / (float)t3f_virtual_display_width;

	/* need to adjust y */
	if(r > vr)
	{
		app->display_offset_x = 0;
		w = tw;
		app->display_offset_y = (th - tw * vr) / 2.0;
		h = tw * vr;
	}
	else
	{
		app->display_offset_x = (tw - th / vr) / 2.0;
		w = th / vr;
		app->display_offset_y = 0;
		h = th;
	}
	app->display_scale_x = w / (float)t3f_virtual_display_width;
	app->display_scale_y = h / (float)(t3f_virtual_display_height - 16.0);
	al_build_transform(&app->transform, app->display_offset_x, app->display_offset_y, app->display_scale_x, app->display_scale_y, 0.0);
}

void crsed_draw_polygon(void * data, ALLEGRO_BITMAP * bp, POLY po, int cx, int cy, ALLEGRO_COLOR c1, ALLEGRO_COLOR c2)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	int i;
	int o = app->zoom / 2;

	for(i = 0; i < po.num_vertices; i += 2)
	{
		al_draw_line(po.p[i].x, po.p[i].y, po.p[i + 1].x, po.p[i + 1].y, c1, 1.0);
	}
	for(i = 0; i < po.num_vertices - 1; i++)
	{
		al_draw_line((cx + po.p[i].x) * app->zoom - app->ox + o, (cy + po.p[i].y) * app->zoom - app->oy + o, (cx + po.p[i+1].x) * app->zoom - app->ox + o, (cy + po.p[i + 1].y) * app->zoom - app->oy + o, c2, 1.0);
	}
}

void draw_circle(void * data, ALLEGRO_BITMAP * bp, int x, int y, int r, ALLEGRO_COLOR color)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	int o = app->zoom / 2;

	al_draw_circle(x * app->zoom - app->ox + o, y * app->zoom - app->oy + o, r * app->zoom, color, 1.0 / app->display_scale_x);
}

void draw_filled_circle(void * data, ALLEGRO_BITMAP * bp, int x, int y, int r, ALLEGRO_COLOR color)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	int o = app->zoom / 2;

	al_draw_filled_circle(x * app->zoom - app->ox + o, y * app->zoom - app->oy + o, r * app->zoom, color);
}

void draw_line(void * data, ALLEGRO_BITMAP * bp, int x1, int y1, int x2, int y2, ALLEGRO_COLOR color)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	int o = app->zoom / 2;

	al_draw_line(x1 * app->zoom - app->ox + o, y1 * app->zoom - app->oy + o, x2 * app->zoom - app->ox + o, y2 * app->zoom - app->oy + o, color, 1.0 / app->display_scale_x);
}

void draw_rectangle(void * data, ALLEGRO_BITMAP * bp, int x1, int y1, int x2, int y2, ALLEGRO_COLOR color)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	int o = app->zoom / 2;

	al_draw_rectangle(x1 * app->zoom - app->ox + o, y1 * app->zoom - app->oy + o, x2 * app->zoom - app->ox + o, y2 * app->zoom - app->oy + o, color, 1.0 / app->display_scale_x);
}

void draw_filled_rectangle(void * data, ALLEGRO_BITMAP * bp, int x1, int y1, int x2, int y2, ALLEGRO_COLOR color)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	int o = app->zoom / 2;

	al_draw_filled_rectangle(x1 * app->zoom - app->ox + o, y1 * app->zoom - app->oy + o, x2 * app->zoom - app->ox + o, y2 * app->zoom - app->oy + o, color);
}

void draw_masked_bitmap(void * data, T3F_ANIMATION * src, ALLEGRO_BITMAP * dest, int xx, int yy, int x, int y, int w, int h)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	t3f_draw_scaled_animation(src, t3f_color_white, 0, x * app->zoom - app->ox, y * app->zoom - app->oy, 0, app->zoom, 0);
}

void draw_masked_bitmap2(void * data, ALLEGRO_BITMAP * src, ALLEGRO_BITMAP * dest, int xx, int yy, int x, int y, int w, int h)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	t3f_draw_scaled_bitmap(src, t3f_color_white, x * app->zoom - app->ox, y * app->zoom - app->oy, 0, app->zoom * w, app->zoom * h, 0);
}

void draw_stretched_sprite(void * data, ALLEGRO_BITMAP * dest, T3F_ANIMATION * src, int x, int y, int w, int h)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	t3f_draw_scaled_animation(src, t3f_color_white, 0, x * app->zoom - app->ox, y * app->zoom - app->oy, 0, ((float)w / (float)src->frame[0]->width) * (float)app->zoom, 0);
}

void draw_vflip_sprite(void * data, ALLEGRO_BITMAP * dest, T3F_ANIMATION * src, int x, int y)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	t3f_draw_scaled_animation(src, t3f_color_white, 0, x * app->zoom - app->ox, y * app->zoom - app->oy, 0, al_itofix(app->zoom), ALLEGRO_FLIP_VERTICAL);
}

void draw_rotated_sprite(void * data, ALLEGRO_BITMAP * dest, T3F_ANIMATION * src, int x, int y, al_fixed angle)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	t3f_draw_rotated_scaled_animation(src, t3f_color_white, 0, src->frame[0]->width / 2, src->frame[0]->height / 2, x * app->zoom - app->ox, y * app->zoom - app->oy, 0, angle, app->zoom, 0);
}

void draw_rotated_vflip_sprite(void * data, ALLEGRO_BITMAP * dest, T3F_ANIMATION * src, int x, int y, al_fixed angle)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	t3f_draw_rotated_scaled_animation(src, t3f_color_white, 0, src->frame[0]->width / 2, src->frame[0]->height / 2, x * app->zoom - app->ox, y * app->zoom - app->oy, 0, angle, app->zoom, ALLEGRO_FLIP_VERTICAL);
}
