#ifndef CRSED_ASSET_H
#define CRSED_ASSET_H

#include "t3f/t3f.h"

#define CRSED_MAX_ASSET_BITMAPS    16
#define CRSED_MAX_ASSET_VERTICES  256

#define CRSED_ASSET_BACKGROUND      0
#define CRSED_ASSET_OBJECT          1
#define CRSED_ASSET_WALL            2
#define CRSED_ASSET_PATTERN         3
#define CRSED_ASSET_VERT            4

#define CRSED_ASSET_DRAW_EDIT       1

typedef struct
{

    float x;
    float y;
    float u;
    float v;
    ALLEGRO_COLOR color;

} CRSED_ASSET_VERTEX;

typedef struct
{

    CRSED_ASSET_VERTEX vertex[CRSED_MAX_ASSET_VERTICES];
    int vertices;

} CRSED_ASSET_POLY;

typedef struct
{

    ALLEGRO_BITMAP * source_bitmap[CRSED_MAX_ASSET_BITMAPS];
    ALLEGRO_BITMAP * bitmap[CRSED_MAX_ASSET_BITMAPS];
    int bitmaps;
    CRSED_ASSET_POLY poly;
    int type;
    int vert_type;
    float x, y, r; // offset and size of vert within bitmap
    ALLEGRO_COLOR default_color;

    char name[256];
    char author[256];
    char comment[1024];

    /* temporary data, do not save */
    float rx, ry;

} CRSED_ASSET;

typedef struct
{

    int asset; // index of asset within currently loaded asset collection
    int asset_type;

    /* object placement */
    float x;
    float y;
    float scale;
    float angle;
    ALLEGRO_COLOR color;

    /* wall placement */
    CRSED_ASSET_VERTEX vertex[CRSED_MAX_ASSET_VERTICES];
    int vertices;

    /* temporary data, do not save */
    float ax, ay;
    float rx, ry;
    bool no_scale;
    bool no_rotate;

} CRSED_ASSET_INSTANCE;

CRSED_ASSET * crsed_create_asset(void);
CRSED_ASSET * crsed_load_asset_f(ALLEGRO_FILE * fp, const char * fn);
CRSED_ASSET * crsed_load_asset(const char * fn);
bool crsed_save_asset_f(CRSED_ASSET * ap, ALLEGRO_FILE * fp);
bool crsed_save_asset(CRSED_ASSET * ap, const char * fn);
void crsed_destroy_asset(CRSED_ASSET * ap);

CRSED_ASSET_INSTANCE * crsed_create_asset_instance(int asset, ALLEGRO_COLOR color, float x, float y, float scale, float angle);
CRSED_ASSET_INSTANCE * crsed_load_asset_instance_f(ALLEGRO_FILE * fp);
bool crsed_save_asset_instance_f(CRSED_ASSET_INSTANCE * ap, ALLEGRO_FILE * fp);
void crsed_destroy_asset_instance(CRSED_ASSET_INSTANCE * ap);

void crsed_render_asset_instance(CRSED_ASSET_INSTANCE * asset_instance, CRSED_ASSET * asset, float ox, float oy, int flags);

#endif
