#include "t3f/t3f.h"
#ifdef ALLEGRO_WINDOWS
	#include "t3f/windows.h"
#endif

#include "instance.h"
#include "main.h"
#include "ui_init.h"
#include "core/resource.h"

void app_load_config(APP_INSTANCE * app)
{
	const char * val;

	val = al_get_config_value(t3f_config, "App Settings", "last_filename");
	if(val)
	{
		strcpy(app->last_filename, val);
	}
	val = al_get_config_value(t3f_config, "App Settings", "last_music_filename");
	if(val)
	{
		strcpy(app->last_music_filename, val);
	}
	val = al_get_config_value(t3f_config, "App Settings", "last_image_filename");
	if(val)
	{
		strcpy(app->last_image_filename, val);
	}
	val = al_get_config_value(t3f_config, "App Settings", "last_hole_filename");
	if(val)
	{
		strcpy(app->last_hole_filename, val);
	}
	val = al_get_config_value(t3f_config, "App Settings", "last_asset_filename");
	if(val)
	{
		strcpy(app->last_asset_filename, val);
	}
	val = al_get_config_value(t3f_config, "App Settings", "music_enabled");
	if(val)
	{
		if(!strcmp(val, "true"))
		{
			app->music_enabled = true;
		}
	}
}

void app_save_config(APP_INSTANCE * app)
{
	al_set_config_value(t3f_config, "App Settings", "last_filename", app->last_filename);
	al_set_config_value(t3f_config, "App Settings", "last_music_filename", app->last_music_filename);
	al_set_config_value(t3f_config, "App Settings", "last_image_filename", app->last_image_filename);
	al_set_config_value(t3f_config, "App Settings", "last_hole_filename", app->last_hole_filename);
	al_set_config_value(t3f_config, "App Settings", "last_asset_filename", app->last_asset_filename);
	al_set_config_value(t3f_config, "App Settings", "music_enabled", app->music_enabled ? "true" : "false");
}

void app_check_args(APP_INSTANCE * app, int argc, char * argv[])
{
	int i;

	for(i = 1; i < argc; i++)
	{
		if(!strcmp(argv[i], "--mode"))
		{
			if(argc > i + 1)
			{
				app->course_mode = atoi(argv[i + 1]);
				if(app->course_mode != 0 && app->course_mode != 1)
				{
					app->course_mode = 0;
				}
			}
		}
	}
}

/* initialize our app, load graphics, etc. */
bool app_initialize(APP_INSTANCE * app, int argc, char * argv[])
{
	/* initialize T3F */
	if(!t3f_initialize(T3F_APP_TITLE, 640, 480 + 16, 60.0, crsed_logic, crsed_render, T3F_DEFAULT | T3F_USE_MENU | T3F_RESIZABLE, app))
	{
		printf("Error initializing T3F\n");
		return false;
	}
	t3f_set_event_handler(crsed_event_proc);
	al_set_new_bitmap_flags(ALLEGRO_NO_PREMULTIPLIED_ALPHA | ALLEGRO_MIN_LINEAR | ALLEGRO_MAG_LINEAR);
	al_set_blender(ALLEGRO_ADD, ALLEGRO_ALPHA, ALLEGRO_INVERSE_ALPHA);
	if(!t3gui_init())
	{
		return false;
	}
	memset(app, 0, sizeof(APP_INSTANCE));
	#ifdef ALLEGRO_WINDOWS
		t3f_windows_text_to_utf8(T3F_APP_COPYRIGHT, app->copyright_message, 256);
	#else
		strcpy(app->copyright_message, T3F_APP_COPYRIGHT);
	#endif
	app_check_args(app, argc, argv);
	app->zoom = 1;
	app->new_poly = -1;
	app->new_point = -1;

	app_load_config(app);

	app->button_theme = t3gui_load_theme("data/button_theme.ini", 0);
	if(!app->button_theme)
	{
		return false;
	}
	app->text_box_theme = t3gui_load_theme("data/listbox_theme.ini", 0);
	if(!app->text_box_theme)
	{
		return false;
	}
	app->asset_theme = t3gui_load_theme("data/button_theme.ini", 0);
	if(!app->button_theme)
	{
		return false;
	}
	app->font = al_load_font("data/Signpgsc.ttf", 16, 0);
	if(!app->font)
	{
		return false;
	}
	ALLEGRO_STATE old_state;
	al_store_state(&old_state, ALLEGRO_STATE_NEW_BITMAP_PARAMETERS);
	al_set_new_bitmap_flags(0); // turn off texture filtering for this load
	app->palette_bitmap = al_load_bitmap("data/palette.png");
	al_restore_state(&old_state);
	if(!app->palette_bitmap)
	{
		return false;
	}
	app->peg_bitmap = al_load_bitmap("data/peg.png");
	if(!app->peg_bitmap)
	{
		return false;
	}
	if(!crsed_setup_menus(app))
	{
		return false;
	}
	if(!crsed_setup_dialogs(app))
	{
		return false;
	}
	t3f_attach_menu(app->main_menu);

	vgolf_register_resource_loaders();
	crsed_set_cursor(app, ALLEGRO_SYSTEM_MOUSE_CURSOR_DEFAULT);
	return true;
}
