#include "t3f/t3f.h"

#include "core/course.h"
#include "core/legacy_course.h"
#include "instance.h"
#include "main.h"
#include "tables.h"
#include "ui.h"
#include "ui_menu_proc.h"
#include "ui_vert.h"

/* for menu items that depend only on a course being loaded */
int menu_file_save_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->course)
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }
    else
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    return 1;
}

/* for updating items in the hole select menu */
int menu_hole_select_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    int i;

    if(!app->course)
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
        return 1;
    }
    /* see if the item is for a hole that is within the current range for the loaded course */
    for(i = 0; i < app->course->num_holes; i++)
    {
        if(item == app->hole_menu_id[i])
        {
            break;
        }
    }
    if(i == app->course->num_holes)
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    else if(item == app->hole_menu_id[app->selected_hole])
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_CHECKED);
    }
    else
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }
    return 1;
}

/* for items that require a course to be loaded and num_holes to be less than max */
int menu_hole_add_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(!app->course)
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    else if(app->course->num_holes < 18)
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }
    else
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    return 1;
}

/* for menu items that make changes to holes (requires a hole to exist) */
int menu_hole_edit_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(!app->course)
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    else if(app->course->num_holes > 0)
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }
    else
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    return 1;
}

int menu_hole_layer_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    int i;

    for(i = 0; i < 3; i++)
    {
        if(app->layer_menu_id[i] == item)
        {
            break;
        }
    }
    if(!app->course)
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    else if(i == app->selected_layer)
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_CHECKED);
    }
    else
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }

    return 1;
}

int menu_hole_image_loaded_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(!app->course)
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    else if(app->course->hole[app->selected_hole].layer_pic[app->selected_layer])
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }
    else
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    return 1;
}

int menu_mode_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    int i;

    for(i = 0; i < 4; i++)
    {
        if(app->mode_menu_id[i] == item)
        {
            break;
        }
    }
    if(!app->course)
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    else if(i == app->mode)
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_CHECKED);
    }
    else
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }
    return 1;
}

/* for menu items that make changes to holes (requires a hole to exist) */
int menu_poly_edit_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->course && app->mode == CRSED_STATE_POLYGONS && app->course->hole[app->selected_hole].bodies[app->selected_layer] > 0)
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }
	else
	{
		t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
	}
    return 1;
}

/* for menu items that make changes to objects */
int menu_object_edit_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->course && app->mode == CRSED_STATE_OBJECTS && app->course->hole[app->selected_hole].num_verts > 0)
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }
	else
	{
		t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
	}
    return 1;
}

/* for menu items that make changes to objects */
int menu_object_move_up_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->course && app->mode == CRSED_STATE_OBJECTS && app->selected_vert < app->course->hole[app->selected_hole].num_verts - 1)
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }
	else
	{
		t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
	}
    return 1;
}

/* for menu items that make changes to objects */
int menu_object_move_down_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->course && app->mode == CRSED_STATE_OBJECTS && app->selected_vert > 0)
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }
	else
	{
		t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
	}
    return 1;
}

int menu_object_type_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    int i;

    for(i = 0; i < MAX_OBJECT_TYPES; i++)
    {
        if(crsed_object_type[i] == app->selected_vert_type)
        {
            break;
        }
    }
    if(!app->course || app->mode != CRSED_STATE_OBJECTS)
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    else if(item == app->object_menu_id[i])
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_CHECKED);
    }
    else
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }
    return 1;
}

int menu_object_properties_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    int i;

    for(i = 0; i < 10; i++)
    {
        t3f_set_menu_item_flags(menu, app->object_properties_menu_id[i], ALLEGRO_MENU_ITEM_DISABLED);
    }
    if(app->mode != CRSED_STATE_OBJECTS)
    {
        return 1;
    }
    switch(app->course->hole[app->selected_hole].vert[app->selected_vert].flag)
    {
        case VERT_FLAG_HOLE:
        {
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_COLOR], 0);
            if(app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[1] == 254)
            {
                t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_ANGLE], 0);
            }
            break;
        }
        case VERT_FLAG_PIPE_ENTER:
        {
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_CONNECT], 0);
            if(app->course->hole[app->selected_hole].vert[app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[1]].flag == VERT_FLAG_PIPE_EXIT)
            {
                t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_ANGLE], 0);
                t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_TIMER], 0);
            }
            break;
        }
        case VERT_FLAG_POLY_MORPH:
        {
//            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_COLOR], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_CONNECT], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_SPEED], 0);
            break;
        }
        case VERT_FLAG_POLY_MOVE_HORIZ:
        case VERT_FLAG_POLY_MOVE_VERT:
        {
//            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_COLOR], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_CONNECT], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_SPEED], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_SIZE], 0);
            break;
        }
        case VERT_FLAG_CLOUD_HORIZ:
        case VERT_FLAG_CLOUD_VERT:
        {
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_COLOR], 0);
            break;
        }
        case VERT_FLAG_LASER_SEND:
        {
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_COLOR], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_CONNECT], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_TIMER], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_TIMER2], 0);
            break;
        }
        case VERT_FLAG_TELEPORT:
        {
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_CONNECT], 0);
            if(app->course->hole[app->selected_hole].vert[app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[0]].flag == VERT_FLAG_TELEPORT)
            {
                t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_ANGLE], 0);
            }
            break;
        }
        case VERT_FLAG_BLINKER:
        {
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_COLOR], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_TIMER], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_SET_TYPE], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_TIMER2], 0);
            break;
        }
        case VERT_FLAG_POLY_DOOR_HORIZ:
        case VERT_FLAG_POLY_DOOR_VERT:
        {
//            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_COLOR], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_CONNECT], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_SPEED], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_TIMER], 0);
            break;
        }
        case VERT_FLAG_CONVEYOR:
        {
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_SPEED], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_CONNECT], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_DIRECTION], 0);
            break;
        }
        case VERT_FLAG_POLY_MOTION_AREA:
        {
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_CONNECT], 0);
            break;
        }
        case VERT_FLAG_LAYER_TRANSFER:
        {
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_CONNECT], 0);
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_SET_TARGET_LAYER], 0);
            break;
        }
        case VERT_FLAG_TRANSLUCENCY:
        {
            t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_CONNECT], 0);
        }
    }
    t3f_set_menu_item_flags(menu, app->object_properties_menu_id[CRSED_GUI_OBJECT_PROPERTIES_SET_LAYER], 0);
    return 1;
}

int menu_image_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    int i;

    for(i = 0; i < MAX_IMAGE_TYPES; i++)
    {
        if(i == app->selected_image)
        {
            break;
        }
    }
    if(!app->course || app->mode != CRSED_STATE_IMAGES)
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    else if(item == app->image_menu_id[i])
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_CHECKED);
    }
    else
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }
    return 1;
}

int menu_build_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(!app->course || app->mode != CRSED_STATE_BUILD)
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    else
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }
    return 1;
}

int menu_music_remove_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->course && strlen(app->music_filename))
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }
    else
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_DISABLED);
    }
    return 1;
}

int menu_music_listen_update_proc(ALLEGRO_MENU * menu, int item, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    if(app->music_enabled)
    {
        t3f_set_menu_item_flags(menu, item, ALLEGRO_MENU_ITEM_CHECKED);
    }
    else
    {
        t3f_set_menu_item_flags(menu, item, 0);
    }
    return 1;
}

static char rfn[1024] = {0};

const char * crsed_get_music(char * fn)
{
	char * type[6] = {".xm", ".it", ".s3m", ".mod", ".ogg", ".wav"};
	int types = 6;
	int i = 0;
	ALLEGRO_PATH * pp;

	pp = al_create_path(fn);
	if(pp)
	{
		for(i = 0; i < types; i++)
		{
			al_set_path_extension(pp, type[i]);
			if(al_filename_exists(al_path_cstr(pp, '/')))
			{
				sprintf(rfn, "%s", al_path_cstr(pp, '/'));
				break;
			}
		}
		al_destroy_path(pp);
	}
	if(i < types)
	{
		return rfn;
	}
	return NULL;
}

static char * ncd_file_select(int type, const char * initial, const char * text, const char * filter)
{
	ALLEGRO_FILECHOOSER * fc;

	fc = al_create_native_file_dialog(initial, text, filter, type == 0 ? ALLEGRO_FILECHOOSER_FILE_MUST_EXIST : ALLEGRO_FILECHOOSER_SAVE);
	if(!fc)
	{
		goto fail;
	}
	if(!al_show_native_file_dialog(al_get_current_display(), fc))
	{
		goto fail;
	}
	if(!al_get_native_file_dialog_count(fc))
	{
		goto fail;
	}
	strcpy(rfn, al_get_native_file_dialog_path(fc, 0));
	return rfn;

	fail:
	{
		if(fc)
		{
			al_destroy_native_file_dialog(fc);
		}
		return NULL;
	}
}

static bool get_new_palette(void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_STATE old_state;
    ALLEGRO_TRANSFORM identity;
    ALLEGRO_BITMAP * bp;
    int i, j;

    al_store_state(&old_state, ALLEGRO_STATE_TARGET_BITMAP | ALLEGRO_STATE_TRANSFORM | ALLEGRO_STATE_NEW_BITMAP_PARAMETERS);
    al_set_new_bitmap_flags(0);
    bp = al_create_bitmap(16, 16);
    if(bp)
    {
        al_set_target_bitmap(bp);
        al_identity_transform(&identity);
        al_use_transform(&identity);
        al_lock_bitmap(bp, ALLEGRO_PIXEL_FORMAT_ANY, ALLEGRO_LOCK_WRITEONLY);
        for(i = 0; i < 16; i++)
        {
            for(j = 0; j < 16; j++)
            {
                al_put_pixel(j, i, al_map_rgb(app->course->palette.entry[i * 16 + j].r, app->course->palette.entry[i * 16 + j].g, app->course->palette.entry[i * 16 + j].b));
            }
        }
        al_unlock_bitmap(bp);
        al_restore_state(&old_state);
        if(app->palette_bitmap)
        {
            al_destroy_bitmap(app->palette_bitmap);
        }
        app->palette_bitmap = bp;
        app->palette_element->dp = bp;
        return true;
    }
    return false;
}

void crsed_post_unsaved_dialog(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	int i, j;

	if(app->unsaved_save)
	{
		crsed_menu_file_save(id, data);
	}
	switch(app->unsaved_type)
	{
		case 0:
		{
			if(app->course)
			{
				vgolf_destroy_course(app->course);
			}
			app->course = vgolf_load_course("data/template.vdc", 0, VGOLF_COURSE_LOAD_ALL | VGOLF_COURSE_LOAD_EDIT, NULL);
            if(app->course)
            {
                app->course->image_bank.ball_pic = t3f_load_animation("data/ball.t3a");
			    app->selected_hole = 0;
			    app->selected_poly = 0;
		        app->selected_vert = 0;
                app->selected_point = 0;
                app->selected_image = 0;
                app->selected_layer = 0;
                app->mode = CRSED_STATE_POLYGONS;
                app->changes = 0;
                crsed_vert_update_live_hole(data);
                get_new_palette(data);
                strcpy(app->loaded_filename, "");
                crsed_fix_window_title(data);
                t3f_refresh_menus();
            }
			break;
		}
		case 1:
		{
			const char * returnedfn = NULL;

			returnedfn = ncd_file_select(0, app->last_filename, "Select Course File", "*.crs;*.vdc");
			if(returnedfn)
			{
				if(app->course)
				{
					vgolf_destroy_course(app->course);
				}
				t3f_enable_menus(false);
				app->course = vgolf_load_course(returnedfn, app->course_mode, VGOLF_COURSE_LOAD_ALL | VGOLF_COURSE_LOAD_EDIT, NULL);
				if(!app->course)
				{
					app->course = vgolf_load_legacy_course(returnedfn, 0, VGOLF_COURSE_LOAD_ALL | VGOLF_COURSE_LOAD_EDIT, NULL);
				}
				t3f_process_events();
				t3f_enable_menus(true);
				if(app->course)
				{
					strcpy(app->loaded_filename, returnedfn);
					strcpy(app->last_filename, returnedfn);
					returnedfn = crsed_get_music(app->loaded_filename);
					if(returnedfn)
					{
                        strcpy(app->music_filename, returnedfn);
                        if(app->music_enabled)
                        {
                            t3f_play_music(returnedfn);
                        }
					}
					app->course->image_bank.ball_pic = t3f_load_animation("data/ball.t3a");
					for(i = 0; i < app->course->num_holes; i++)
					{
						for(j = 0; j < app->course->hole[i].num_verts; j++)
						{
							if(app->course->hole[i].vert[j].flag == VERT_FLAG_TREE_0 && app->course->hole[i].vert[j].radius <= 0)
							{
								app->course->hole[i].vert[j].radius = app->course->image_bank.object_pic[0][0]->frame[0]->width / 2;
							}
							else if(app->course->hole[i].vert[j].flag == VERT_FLAG_TREE_1 && app->course->hole[i].vert[j].radius <= 0)
							{
								app->course->hole[i].vert[j].radius = app->course->image_bank.object_pic[1][0]->frame[0]->width / 2;
							}
						}
					}
					app->selected_hole = 0;
					app->selected_image = 0;
					app->selected_poly = 0;
					app->selected_point = 0;
					app->selected_vert = 0;
					app->selected_layer = 0;
					app->mode = CRSED_STATE_POLYGONS;
					app->changes = 0;
					crsed_vert_update_live_hole(data);
                    get_new_palette(data);
					crsed_fix_window_title(data);
				}
				t3f_refresh_menus();
			}
			break;
		}
		case 2:
		{
			app_save_config(app);
			t3f_exit();
			break;
		}
	}
}

int crsed_menu_file_new(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	app->unsaved_type = 0;
	if(app->course && app->changes)
	{
		t3f_enable_menus(false);
		t3gui_show_dialog(app->unsaved_dialog, t3f_queue, 0, data);
	}
	else
	{
		crsed_post_unsaved_dialog(id, data);
	}
	return 1;
}

int crsed_menu_file_load(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	app->unsaved_type = 1;
	if(app->course && app->changes)
	{
		t3f_enable_menus(false);
		t3gui_show_dialog(app->unsaved_dialog, t3f_queue, 0, data);
	}
	else
	{
		crsed_post_unsaved_dialog(id, data);
	}
	return 1;
}

static void update_course_palette(void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
    ALLEGRO_COLOR color;
    unsigned char r, g, b;
    int i, j;

    al_lock_bitmap(app->palette_bitmap, ALLEGRO_PIXEL_FORMAT_ANY, ALLEGRO_LOCK_READONLY);
    for(i = 0; i < 16; i++)
    {
        for(j = 0; j < 16; j++)
        {
            color = al_get_pixel(app->palette_bitmap, j, i);
            al_unmap_rgb(color, &r, &g, &b);
            app->course->palette.entry[i * 16 + j].r = r;
            app->course->palette.entry[i * 16 + j].g = g;
            app->course->palette.entry[i * 16 + j].b = b;
        }
    }
    al_unlock_bitmap(app->palette_bitmap);
}

int crsed_menu_file_load_palette(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
	char * returnedfn = NULL;
    ALLEGRO_BITMAP * bp;

	if(app->course)
	{
		returnedfn = ncd_file_select(0, app->last_image_filename, "Select Palette File", "*.png");
		if(returnedfn)
		{
            ALLEGRO_STATE old_state;
        	al_store_state(&old_state, ALLEGRO_STATE_NEW_BITMAP_PARAMETERS);
        	al_set_new_bitmap_flags(0); // turn off texture filtering for this load
        	bp = al_load_bitmap(returnedfn);
        	al_restore_state(&old_state);
            if(bp)
            {
                if(app->palette_bitmap)
                {
                    al_destroy_bitmap(app->palette_bitmap);
                }
                app->palette_bitmap = bp;
                app->palette_element->dp = bp;
                update_course_palette(data);
                strcpy(app->last_image_filename, returnedfn);
            }
		}
	}
	return 1;
}

int crsed_menu_file_save_palette(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
	char * returnedfn = NULL;
    ALLEGRO_PATH * path;

    if(app->course)
    {
        returnedfn = ncd_file_select(1, app->last_image_filename, "Select Palette File", "*.png");
		if(returnedfn)
		{
            path = al_create_path(returnedfn);
            if(path)
            {
                al_set_path_extension(path, ".png");
                al_save_bitmap(al_path_cstr(path, '/'), app->palette_bitmap);
                strcpy(app->last_image_filename, al_path_cstr(path, '/'));
                al_destroy_path(path);
            }
		}
    }
    return 1;
}

int crsed_menu_file_course_tags(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	sprintf(crsed_edit_text[0], "%s", app->course->name);
	sprintf(crsed_edit_text[1], "%s", app->course->author);
	sprintf(crsed_edit_text[2], "%s", app->course->comment);
	t3f_enable_menus(false);
	t3gui_show_dialog(app->tags_dialog, t3f_queue, 0, data);
	return 1;
}

int crsed_menu_file_save(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	ALLEGRO_PATH * pp;
	ALLEGRO_PATH * mpp;

	if(app->course)
	{
		if(!strcmp(app->loaded_filename, ""))
		{
			crsed_menu_file_save_as(id, data);
		}
		else
		{
			pp = al_create_path(app->loaded_filename);
			if(pp)
			{
				al_set_path_extension(pp, ".vdc");
				strcpy(app->loaded_filename, al_path_cstr(pp, '/'));
				t3f_enable_menus(false);

                al_set_path_extension(pp, ".vdc");
				vgolf_save_course(app->course, app->loaded_filename, app->course_mode);
				t3f_process_events();
				t3f_enable_menus(true);

				/* copy music file if necessary */
				if(!crsed_get_music(app->loaded_filename) && strlen(app->music_filename))
				{
					mpp = al_create_path(app->music_filename);
					if(mpp)
					{
						al_set_path_extension(pp, al_get_path_extension(mpp));
						t3f_copy_file(app->music_filename, al_path_cstr(pp, '/'));
						al_destroy_path(mpp);
						strcpy(app->music_filename, al_path_cstr(pp, '/'));
					}
				}
				al_destroy_path(pp);

				app->changes = 0;
				crsed_fix_window_title(data);
			}
		}
	}
	return 1;
}

int crsed_menu_file_save_as(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	ALLEGRO_PATH * pp;
	char * returnedfn = NULL;

	if(app->course)
	{
		pp = al_create_path(app->loaded_filename);
		if(pp)
		{
			al_set_path_extension(pp, ".vdc");
			strcpy(app->loaded_filename, al_path_cstr(pp, '/'));
			al_destroy_path(pp);
			returnedfn = ncd_file_select(1, app->loaded_filename, "Save Course As", "*.vdc");
			if(returnedfn)
			{
				pp = al_create_path(returnedfn);
				if(pp)
				{
					al_set_path_extension(pp, ".vdc");
					strcpy(app->loaded_filename, al_path_cstr(pp, '/'));
					strcpy(app->last_filename, app->loaded_filename);
					al_destroy_path(pp);
					crsed_menu_file_save(id, data);
				}
			}
		}
	}
	return 1;
}

int crsed_menu_file_package(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	ALLEGRO_PATH * pp;
	ALLEGRO_PATH * mpp;
    ALLEGRO_PATH * cpp;
    unsigned long course_checksum;
    char buf[128];
    char sys_command[1024];
    char * cwd;

	if(app->course)
	{
        if(app->changes)
        {
            crsed_menu_file_save(id, data);
        }
		pp = al_create_path(app->loaded_filename);
		if(pp)
		{
            cpp = al_clone_path(pp);
            if(cpp)
            {
                if(strlen(app->music_filename) > 0)
                {
                    mpp = al_create_path(app->music_filename);
                    if(mpp)
                    {
                        t3f_enable_menus(false);
                        course_checksum = t3f_checksum_file(al_path_cstr(pp, '/'));
                        al_set_path_extension(pp, "");
                        sprintf(sys_command, "zip -9 %s.zip %s.vdc %s.gsd %s.ghd %s%s", al_get_path_filename(pp), al_get_path_filename(pp), al_get_path_filename(pp), al_get_path_filename(pp), al_get_path_filename(pp), al_get_path_extension(mpp));
                        cwd = al_get_current_directory();
                        if(cwd)
                        {
                            al_set_path_filename(cpp, "");
                            if(al_change_directory(al_path_cstr(cpp, '/')))
                            {
                                system(sys_command);
                                sprintf(buf, ".%lu.vdci", course_checksum);
                                al_set_path_extension(pp, buf);
                                vgolf_save_course_info(app->course, al_path_cstr(pp, '/'));
                                al_change_directory(cwd);
                            }
                            free(cwd);
                        }
                        t3f_process_events();
                        t3f_enable_menus(true);
                        al_destroy_path(mpp);
                    }
                }
                al_destroy_path(cpp);
            }
            al_destroy_path(pp);
        }
    }
	return 1;
}

int crsed_menu_file_exit(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	app->unsaved_type = 2;
	if(app->changes)
	{
		t3f_enable_menus(false);
		t3gui_show_dialog(app->unsaved_dialog, t3f_queue, 0, data);
	}
	else
	{
		crsed_post_unsaved_dialog(id, data);
	}
	return 1;
}

int crsed_menu_mode_polygons(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	app->mode = CRSED_STATE_POLYGONS;
    crsed_vert_update_live_hole(data);
	t3f_refresh_menus();
	return 1;
}

int crsed_menu_mode_objects(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	app->mode = CRSED_STATE_OBJECTS;
    crsed_vert_update_live_hole(data);
	t3f_refresh_menus();
	return 1;
}

int crsed_menu_mode_images(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	app->mode = CRSED_STATE_IMAGES;
	t3f_refresh_menus();
	return 1;
}

static void crsed_menu_hole_select_helper(void * data, int hole)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	app->selected_hole = hole;
	app->selected_poly = 0;
	app->selected_point = 0;
	app->selected_vert = 0;
	app->selected_layer = 0;
	crsed_vert_update_live_hole(data);
	t3f_refresh_menus();
}

int crsed_menu_hole_select_0(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 0);
	return 1;
}

int crsed_menu_hole_select_1(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 1);
	return 1;
}

int crsed_menu_hole_select_2(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 2);
	return 1;
}

int crsed_menu_hole_select_3(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 3);
	return 1;
}

int crsed_menu_hole_select_4(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 4);
	return 1;
}

int crsed_menu_hole_select_5(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 5);
	return 1;
}

int crsed_menu_hole_select_6(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 6);
	return 1;
}

int crsed_menu_hole_select_7(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 7);
	return 1;
}

int crsed_menu_hole_select_8(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 8);
	return 1;
}

int crsed_menu_hole_select_9(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 9);
	return 1;
}

int crsed_menu_hole_select_10(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 10);
	return 1;
}

int crsed_menu_hole_select_11(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 11);
	return 1;
}

int crsed_menu_hole_select_12(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 12);
	return 1;
}

int crsed_menu_hole_select_13(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 13);
	return 1;
}

int crsed_menu_hole_select_14(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 14);
	return 1;
}

int crsed_menu_hole_select_15(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 15);
	return 1;
}

int crsed_menu_hole_select_16(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 16);
	return 1;
}

int crsed_menu_hole_select_17(int id, void * data)
{
	crsed_menu_hole_select_helper(data, 17);
	return 1;
}

int crsed_menu_hole_add(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	if(app->course->num_holes < 18)
	{
		memset(&app->course->hole[app->course->num_holes], 0, sizeof(HOLE));
		app->course->hole[app->course->num_holes].s.x = 320;
		app->course->hole[app->course->num_holes].s.y = 240;
        app->course->hole[app->course->num_holes].body_layers = MAX_HOLE_LAYERS;
		app->course->num_holes++;
		crsed_mark_change(data);
		t3f_refresh_menus();
	}
	return 1;
}

int crsed_menu_hole_load_image(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	ALLEGRO_BITMAP * temp_image;
	char * returnedfn;

	if(app->course)
	{
		returnedfn = ncd_file_select(0, app->last_image_filename, "Select Image File", "*.png;*.jpg");
		if(returnedfn)
		{
			temp_image = al_load_bitmap(returnedfn);
			if(temp_image)
			{
				if(app->course->hole[app->selected_hole].layer_pic[app->selected_layer])
				{
					al_destroy_bitmap(app->course->hole[app->selected_hole].layer_pic[app->selected_layer]);
				}
				app->course->hole[app->selected_hole].layer_pic[app->selected_layer] = temp_image;
				strcpy(app->last_image_filename, returnedfn);
				crsed_mark_change(data);
				t3f_refresh_menus();
			}
		}
	}
	return 1;
}

int crsed_menu_hole_save_image(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	char * returnedfn;

	if(app->course)
	{
		returnedfn = ncd_file_select(1, app->last_image_filename, "Save Image File", "*.png");
		if(returnedfn)
		{
			al_save_bitmap(returnedfn, app->course->hole[app->selected_hole].layer_pic[app->selected_layer]);
			strcpy(app->last_image_filename, returnedfn);
		}
	}
	return 1;
}

int crsed_menu_hole_remove_image(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	if(app->course->hole[app->selected_hole].layer_pic[app->selected_layer])
	{
		al_destroy_bitmap(app->course->hole[app->selected_hole].layer_pic[app->selected_layer]);
		app->course->hole[app->selected_hole].layer_pic[app->selected_layer] = NULL;
		crsed_mark_change(data);
		t3f_refresh_menus();
	}
	return 1;
}

int crsed_menu_hole_set_par(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	sprintf(crsed_edit_text[0], "%d", app->course->hole[app->selected_hole].par_score);
	t3f_enable_menus(false);
	t3gui_show_dialog(app->par_dialog, t3f_queue, 0, data);
	return 1;
}

int crsed_menu_hole_import(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	char * returnedfn;
	ALLEGRO_FILE * fp;
	int i;

	if(app->course)
	{
		returnedfn = ncd_file_select(0, app->last_hole_filename, "Import Hole File", "*.vdh");
		if(returnedfn)
		{
			fp = al_fopen(returnedfn, "r");
			if(fp)
			{
				vgolf_load_hole_fp(&app->course->hole[app->selected_hole], fp);
				for(i = 0; i < MAX_HOLE_LAYERS; i++)
				{
					if(al_fgetc(fp))
					{
						app->course->hole[app->selected_hole].layer_pic[i] = t3f_load_bitmap_f(fp);
					}
				}
				al_fclose(fp);
				strcpy(app->last_hole_filename, returnedfn);
				crsed_mark_change(data);
			}
		}
	}
	return 1;
}

int crsed_menu_hole_export(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	char * returnedfn;
	ALLEGRO_FILE * fp;
	int i;

	if(app->course)
	{
		returnedfn = ncd_file_select(1, app->last_hole_filename, "Export Hole File", "vdh");
		if(returnedfn)
		{
			fp = al_fopen(returnedfn, "w");
			if(fp)
			{
				vgolf_save_hole_fp(&app->course->hole[app->selected_hole], fp);
				for(i = 0; i < MAX_HOLE_LAYERS; i++)
				{
					if(app->course->hole[app->selected_hole].layer_pic[i])
					{
						al_fputc(fp, 1);
						t3f_save_bitmap_f(fp, app->course->hole[app->selected_hole].layer_pic[i]);
					}
				}
				strcpy(app->last_hole_filename, returnedfn);
				al_fclose(fp);
			}
		}
	}
	return 1;
}

int crsed_menu_hole_delete(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	int i;

	vgolf_free_hole(&app->course->hole[app->selected_hole]);
	for(i = app->selected_hole; i < app->course->num_holes - 1; i++)
	{
		memcpy(&app->course->hole[i], &app->course->hole[i + 1], sizeof(HOLE));
	}
	app->course->num_holes--;
	if(app->selected_hole >= app->course->num_holes)
	{
		if(app->course->num_holes > 0)
		{
			app->selected_hole = app->course->num_holes - 1;
		}
		else
		{
			app->selected_hole = 0;
		}
	}
	crsed_mark_change(data);
	t3f_refresh_menus();
	return 1;
}

int crsed_menu_poly_layer_0(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	if(app->selected_layer != 0)
	{
		app->selected_poly = 0;
		app->selected_layer = 0;
		t3f_refresh_menus();
	}
	return 1;
}

int crsed_menu_poly_layer_1(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	if(app->selected_layer != 1)
	{
		app->selected_poly = 0;
		app->selected_layer = 1;
		t3f_refresh_menus();
	}
	return 1;
}

int crsed_menu_poly_layer_2(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	if(app->selected_layer != 2)
	{
		app->selected_poly = 0;
		app->selected_layer = 2;
		t3f_refresh_menus();
	}
	return 1;
}

int crsed_menu_poly_invert(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	POLY temp_poly;
	int i;

	memcpy(&temp_poly, &app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly, sizeof(POLY));
	for(i = 0; i < temp_poly.num_vertices; i++)
	{
		app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.p[i] = temp_poly.p[temp_poly.num_vertices - 1 - i];
	}
	crsed_mark_change(data);
	return 1;
}

static void crsed_delete_poly(int id, void * data, int poly)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	int i;

	for(i = poly; i < app->course->hole[app->selected_hole].bodies[app->selected_layer] - 1; i++)
	{
		memcpy(&app->course->hole[app->selected_hole].body[app->selected_layer][i], &app->course->hole[app->selected_hole].body[app->selected_layer][i + 1], sizeof(BODY));
	}
	app->course->hole[app->selected_hole].bodies[app->selected_layer]--;
	if(app->selected_poly >= app->course->hole[app->selected_hole].bodies[app->selected_layer])
	{
		app->selected_poly = app->course->hole[app->selected_hole].bodies[app->selected_layer] - 1;
		if(app->selected_poly < 0)
		{
			app->selected_poly = 0;
			t3f_refresh_menus();
		}
	}
	crsed_mark_change(data);
}

static void crsed_delete_vertex(void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	int i;

    for(i = app->selected_point; i < app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.num_vertices - 1; i++)
    {
        memcpy(&app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.p[i], &app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.p[i + 1], sizeof(VERTEX));
    }
    app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.num_vertices--;
    if(app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.num_vertices < 2)
    {
        crsed_menu_poly_delete(0, data);
    }
    else if(app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.num_vertices > 3)
    {
        if(app->selected_point == 0)
        {
            memcpy(&app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.p[app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.num_vertices - 1], &app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.p[0], sizeof(VERTEX));
        }
        else if(app->selected_point == app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.num_vertices)
        {
            memcpy(&app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.p[0], &app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.p[app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.num_vertices - 1], sizeof(VERTEX));
        }
    }
    else if(app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.num_vertices == 3)
    {
        crsed_menu_poly_delete(0, data);
    }
    if(app->selected_point >= app->course->hole[app->selected_hole].body[app->selected_layer][app->selected_poly].poly.num_vertices)
    {
        app->selected_point--;
    }
    crsed_mark_change(data);
}

int crsed_menu_poly_delete(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	crsed_delete_poly(id, data, app->selected_poly);
	return 1;
}

int crsed_menu_poly_delete_vertex(int id, void * data)
{
    crsed_delete_vertex(data);
    return 1;
}

void crsed_menu_object_type_helper(void * data, int type)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	app->selected_vert_type = type;
	t3f_refresh_menus();
}

int crsed_menu_object_type_TEE(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_TEE);
	return 1;
}

int crsed_menu_object_type_CUP(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_HOLE);
	return 1;
}

int crsed_menu_object_type_TREE_0(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_TREE_0);
	return 1;
}

int crsed_menu_object_type_TREE_1(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_TREE_1);
	return 1;
}

int crsed_menu_object_type_TREE_2(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_TREE_2);
	return 1;
}

int crsed_menu_object_type_TREE_3(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_TREE_3);
	return 1;
}

int crsed_menu_object_type_ANIMAL_0(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_ANIMAL_0);
	return 1;
}

int crsed_menu_object_type_ANIMAL_1(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_ANIMAL_1);
	return 1;
}

int crsed_menu_object_type_ANIMAL_2(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_ANIMAL_2);
	return 1;
}

int crsed_menu_object_type_HAZARD_WATER(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_HAZARD_WATER);
	return 1;
}

int crsed_menu_object_type_HAZARD_SAND(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_HAZARD_SAND);
	return 1;
}

int crsed_menu_object_type_HAZARD_ROUGH(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_HAZARD_ROUGH);
	return 1;
}

int crsed_menu_object_type_HAZARD_SLIME(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_HAZARD_SLIME);
	return 1;
}

int crsed_menu_object_type_PIPE_ENTER(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_PIPE_ENTER);
	return 1;
}

int crsed_menu_object_type_PIPE_EXIT(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_PIPE_EXIT);
	return 1;
}

int crsed_menu_object_type_HILL(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_INDENT_HILL);
	return 1;
}

int crsed_menu_object_type_RUT(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_INDENT_RUT);
	return 1;
}

int crsed_menu_object_type_POLY_MORPH(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_POLY_MORPH);
	return 1;
}

int crsed_menu_object_type_POLY_MOTION_AREA(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_POLY_MOTION_AREA);
	return 1;
}

int crsed_menu_object_type_POLY_MOVE_HORIZ(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_POLY_MOVE_HORIZ);
	return 1;
}

int crsed_menu_object_type_POLY_MOVE_VERT(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_POLY_MOVE_VERT);
	return 1;
}

int crsed_menu_object_type_CLOUD_HORIZ(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_CLOUD_HORIZ);
	return 1;
}

int crsed_menu_object_type_CLOUD_VERT(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_CLOUD_VERT);
	return 1;
}

int crsed_menu_object_type_LASER_SEND(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_LASER_SEND);
	return 1;
}

int crsed_menu_object_type_LASER_RECEIVE(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_LASER_RECV);
	return 1;
}

int crsed_menu_object_type_LASER_RESPAWN(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_LASER_RESPAWN);
	return 1;
}

int crsed_menu_object_type_TELEPORT(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_TELEPORT);
	return 1;
}

int crsed_menu_object_type_CONVEYOR(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_CONVEYOR);
	return 1;
}

int crsed_menu_object_type_LAYER_TRANSFER(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_LAYER_TRANSFER);
	return 1;
}

int crsed_menu_object_type_TRANSLUCENCY(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_TRANSLUCENCY);
	return 1;
}

int crsed_menu_object_type_BLINKER(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_BLINKER);
	return 1;
}

int crsed_menu_object_type_DOOR_HORIZ(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_POLY_DOOR_HORIZ);
	return 1;
}

int crsed_menu_object_type_DOOR_VERTICAL(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_POLY_DOOR_VERT);
	return 1;
}

int crsed_menu_object_type_DOOR_RESPAWN(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_SMASH_RESPAWN);
	return 1;
}

int crsed_menu_object_type_BUBBLE_HORIZ(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_BUBBLE_HORIZ);
	return 1;
}

int crsed_menu_object_type_BUBBLE_VERT(int id, void * data)
{
	crsed_menu_object_type_helper(data, VERT_FLAG_BUBBLE_VERT);
	return 1;
}

int crsed_menu_object_properties_color(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	switch(app->course->hole[app->selected_hole].vert[app->selected_vert].flag)
	{
		case VERT_FLAG_HOLE:
		{
			app->palette_dialog->element[1].d1 = app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[1];
			break;
		}
		case VERT_FLAG_POLY_MORPH:
		case VERT_FLAG_POLY_MOVE_HORIZ:
		case VERT_FLAG_POLY_MOVE_VERT:
		{
			app->palette_dialog->element[1].d1 = app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[3];
			break;
		}
		case VERT_FLAG_LASER_SEND:
		{
			app->palette_dialog->element[1].d1 = app->course->hole[app->selected_hole].vert[app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[5]].ref_no[1];
			break;
		}
		case VERT_FLAG_BLINKER:
		case VERT_FLAG_POLY_DOOR_HORIZ:
		case VERT_FLAG_POLY_DOOR_VERT:
		case VERT_FLAG_CLOUD_HORIZ:
		case VERT_FLAG_CLOUD_VERT:
		{
			app->palette_dialog->element[1].d1 = app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[5];
			break;
		}
	}
	t3f_enable_menus(false);
	t3gui_show_dialog(app->palette_dialog, t3f_queue, 0, data);
	return 1;
}

int crsed_menu_object_properties_connect(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	app->vert_mode = 1;
	app->connect_count = 0;
	crsed_mark_change(data);
	return 1;
}

int crsed_menu_object_properties_set_type(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

	sprintf(crsed_edit_text[0], "%d", app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[4]);
	t3f_enable_menus(false);
	t3gui_show_dialog(app->type_dialog, t3f_queue, 0, data);

    return 1;
}

int crsed_menu_object_properties_set_angle(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	app->vert_mode = 2;
	crsed_mark_change(data);
	return 1;
}

int crsed_menu_object_properties_set_timer(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	switch(app->course->hole[app->selected_hole].vert[app->selected_vert].flag)
	{
		case VERT_FLAG_PIPE_ENTER:
		{
			sprintf(crsed_edit_text[0], "%d", app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[3]);
			break;
		}
		case VERT_FLAG_BLINKER:
		case VERT_FLAG_LASER_SEND:
		{
			sprintf(crsed_edit_text[0], "%d", app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[2]);
			break;
		}
		case VERT_FLAG_POLY_DOOR_HORIZ:
		case VERT_FLAG_POLY_DOOR_VERT:
		{
			sprintf(crsed_edit_text[0], "%d", app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[2]);
			break;
		}
	}
	t3f_enable_menus(false);
	t3gui_show_dialog(app->timer_dialog, t3f_queue, 0, data);
	return 1;
}

int crsed_menu_object_properties_set_timer2(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	switch(app->course->hole[app->selected_hole].vert[app->selected_vert].flag)
	{
		case VERT_FLAG_BLINKER:
		case VERT_FLAG_LASER_SEND:
		{
			sprintf(crsed_edit_text[0], "%d", app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[3]);
			break;
		}
	}
	t3f_enable_menus(false);
	t3gui_show_dialog(app->timer2_dialog, t3f_queue, 0, data);
	return 1;
}

int crsed_menu_object_properties_set_speed(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	switch(app->course->hole[app->selected_hole].vert[app->selected_vert].flag)
	{
		case VERT_FLAG_POLY_MORPH:
		{
			sprintf(crsed_edit_text[0], "%d", app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[2]);
			break;
		}
		case VERT_FLAG_POLY_MOVE_HORIZ:
		case VERT_FLAG_POLY_MOVE_VERT:
		{
			sprintf(crsed_edit_text[0], "%d", app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[5]);
			break;
		}
		case VERT_FLAG_POLY_DOOR_HORIZ:
		case VERT_FLAG_POLY_DOOR_VERT:
		{
			sprintf(crsed_edit_text[0], "%d", app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[3]);
			break;
		}
		case VERT_FLAG_CONVEYOR:
		{
			sprintf(crsed_edit_text[0], "%d", app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[2]);
			break;
		}
		case VERT_FLAG_ANIMAL_0:
		case VERT_FLAG_ANIMAL_1:
		{
			sprintf(crsed_edit_text[0], "%d", app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[1]);
			sprintf(crsed_edit_text[1], "%d", app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[2]);
			break;
		}
	}
	t3f_enable_menus(false);
	t3gui_show_dialog(app->speed_dialog, t3f_queue, 0, data);
	return 1;
}

int crsed_menu_object_properties_set_size(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	sprintf(crsed_edit_text[0], "%d", app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[4] - app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[2]);
	t3f_enable_menus(false);
	t3gui_show_dialog(app->size_dialog, t3f_queue, 0, data);
	return 1;
}

int crsed_menu_object_properties_set_direction(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	app->direction_dialog->element[1].flags = 0;
	app->direction_dialog->element[2].flags = 0;
	app->direction_dialog->element[3].flags = 0;
	app->direction_dialog->element[4].flags = 0;
	switch(app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[0])
	{
		case 0:
		{
			app->direction_dialog->element[1].flags = D_SELECTED;
			break;
		}
		case 1:
		{
			app->direction_dialog->element[2].flags = D_SELECTED;
			break;
		}
		case 2:
		{
			app->direction_dialog->element[3].flags = D_SELECTED;
			break;
		}
		default:
		{
			app->direction_dialog->element[4].flags = D_SELECTED;
			break;
		}
	}
	t3f_enable_menus(false);
	t3gui_show_dialog(app->direction_dialog, t3f_queue, 0, data);
	return 1;
}

int crsed_menu_object_properties_set_target_layer(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	sprintf(crsed_edit_text[0], "%d", app->course->hole[app->selected_hole].vert[app->selected_vert].ref_no[1]);
	t3f_enable_menus(false);
	t3gui_show_dialog(app->target_layer_dialog, t3f_queue, 0, data);
	return 1;
}

int crsed_menu_object_properties_set_layer(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	sprintf(crsed_edit_text[0], "%d", app->course->hole[app->selected_hole].vert[app->selected_vert].layer);
	t3f_enable_menus(false);
	t3gui_show_dialog(app->layer_dialog, t3f_queue, 0, data);
	return 1;
}

static void crsed_delete_vert(void * data, int vert)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	int i;

	for(i = vert; i < app->course->hole[app->selected_hole].num_verts - 1; i++)
	{
		memcpy(&app->course->hole[app->selected_hole].vert[i], &app->course->hole[app->selected_hole].vert[i + 1], sizeof(VERT));
	}
	app->course->hole[app->selected_hole].num_verts--;
	for(i = 0; i < app->course->hole[app->selected_hole].num_verts; i++)
	{
		if(app->course->hole[app->selected_hole].vert[i].flag == VERT_FLAG_TELEPORT)
		{
			if(app->course->hole[app->selected_hole].vert[i].ref_no[0] > vert)
			{
				app->course->hole[app->selected_hole].vert[i].ref_no[0]--;
			}
			else if(app->course->hole[app->selected_hole].vert[i].ref_no[0] >= app->course->hole[app->selected_hole].num_verts)
			{
				app->course->hole[app->selected_hole].vert[i].ref_no[0] = 0;
			}
		}
	}
	if(app->selected_vert >= app->course->hole[app->selected_hole].num_verts)
	{
		app->selected_vert = app->course->hole[app->selected_hole].num_verts - 1;
		if(app->selected_vert < 0)
		{
			app->selected_vert = 0;
			t3f_refresh_menus();
		}
	}
	crsed_vert_update_live_hole(data);
	crsed_mark_change(data);
}

int crsed_menu_object_move_up(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	VERT vert;

	if(app->selected_vert < app->course->hole[app->selected_hole].num_verts - 1)
	{
		memcpy(&vert, &app->course->hole[app->selected_hole].vert[app->selected_vert], sizeof(VERT));
		memcpy(&app->course->hole[app->selected_hole].vert[app->selected_vert], &app->course->hole[app->selected_hole].vert[app->selected_vert + 1], sizeof(VERT));
		memcpy(&app->course->hole[app->selected_hole].vert[app->selected_vert + 1], &vert, sizeof(VERT));
		app->selected_vert++;
		crsed_vert_update_live_hole(data);
		crsed_mark_change(data);
		t3f_refresh_menus();
	}
	return 1;
}

int crsed_menu_object_move_down(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	VERT vert;

	if(app->selected_vert > 0)
	{
		memcpy(&vert, &app->course->hole[app->selected_hole].vert[app->selected_vert], sizeof(VERT));
		memcpy(&app->course->hole[app->selected_hole].vert[app->selected_vert], &app->course->hole[app->selected_hole].vert[app->selected_vert - 1], sizeof(VERT));
		memcpy(&app->course->hole[app->selected_hole].vert[app->selected_vert - 1], &vert, sizeof(VERT));
		app->selected_vert--;
		crsed_vert_update_live_hole(data);
		crsed_mark_change(data);
		t3f_refresh_menus();
	}
	return 1;
}

int crsed_menu_object_delete(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	crsed_delete_vert(data, app->selected_vert);
	return 1;
}

static void crsed_select_image(void * data, int image)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	app->selected_image = image;
	t3f_refresh_menus();
}

int crsed_menu_image_type_title(int id, void * data)
{
	crsed_select_image(data, 0);
	return 1;
}

int crsed_menu_image_type_scoreboard(int id, void * data)
{
	crsed_select_image(data, 1);
	return 1;
}

int crsed_menu_image_type_end_bg(int id, void * data)
{
	crsed_select_image(data, 2);
	return 1;
}

int crsed_menu_image_type_end_podium(int id, void * data)
{
	crsed_select_image(data, 3);
	return 1;
}

int crsed_menu_image_type_end_cloud_1(int id, void * data)
{
	crsed_select_image(data, 4);
	return 1;
}

int crsed_menu_image_type_end_cloud_2(int id, void * data)
{
	crsed_select_image(data, 5);
	return 1;
}

int crsed_menu_image_type_1st_trophy(int id, void * data)
{
	crsed_select_image(data, 6);
	return 1;
}

int crsed_menu_image_type_2nd_trophy(int id, void * data)
{
	crsed_select_image(data, 7);
	return 1;
}

int crsed_menu_image_type_3rd_trophy(int id, void * data)
{
	crsed_select_image(data, 8);
	return 1;
}

int crsed_menu_image_type_object_hole(int id, void * data)
{
	crsed_select_image(data, 9);
	return 1;
}

int crsed_menu_image_type_object_hole_sunk(int id, void * data)
{
	crsed_select_image(data, 10);
	return 1;
}

int crsed_menu_image_type_object_tree_1(int id, void * data)
{
	crsed_select_image(data, 11);
	return 1;
}

int crsed_menu_image_type_object_tree_2(int id, void * data)
{
	crsed_select_image(data, 12);
	return 1;
}

int crsed_menu_image_type_object_laser_1(int id, void * data)
{
	crsed_select_image(data, 13);
	return 1;
}

int crsed_menu_image_type_object_laser_2(int id, void * data)
{
	crsed_select_image(data, 14);
	return 1;
}

int crsed_menu_image_type_object_teleport_1(int id, void * data)
{
	crsed_select_image(data, 15);
	return 1;
}

int crsed_menu_image_type_object_teleport_2(int id, void * data)
{
	crsed_select_image(data, 16);
	return 1;
}

int crsed_menu_image_type_object_conveyor_1(int id, void * data)
{
	crsed_select_image(data, 17);
	return 1;
}

int crsed_menu_image_type_object_conveyor_2(int id, void * data)
{
	crsed_select_image(data, 18);
	return 1;
}

int crsed_menu_image_type_object_conveyor_3(int id, void * data)
{
	crsed_select_image(data, 19);
	return 1;
}

int crsed_menu_image_type_object_conveyor_4(int id, void * data)
{
	crsed_select_image(data, 20);
	return 1;
}

int crsed_menu_image_type_object_animal_1_1(int id, void * data)
{
	crsed_select_image(data, 21);
	return 1;
}

int crsed_menu_image_type_object_animal_1_2(int id, void * data)
{
	crsed_select_image(data, 22);
	return 1;
}

int crsed_menu_image_type_object_animal_1_3(int id, void * data)
{
	crsed_select_image(data, 23);
	return 1;
}

int crsed_menu_image_type_object_animal_1_4(int id, void * data)
{
	crsed_select_image(data, 24);
	return 1;
}

int crsed_menu_image_type_object_animal_2_1(int id, void * data)
{
	crsed_select_image(data, 25);
	return 1;
}

int crsed_menu_image_type_object_animal_2_2(int id, void * data)
{
	crsed_select_image(data, 26);
	return 1;
}

int crsed_menu_image_type_object_animal_2_3(int id, void * data)
{
	crsed_select_image(data, 27);
	return 1;
}

int crsed_menu_image_type_object_animal_2_4(int id, void * data)
{
	crsed_select_image(data, 28);
	return 1;
}

int crsed_menu_image_load(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	T3F_ANIMATION * temp_image;
	char * returnedfn;

	if(app->course)
	{
		returnedfn = ncd_file_select(0, app->last_image_filename, "Select Image File", "*.png;*.jpg;*.t3a");
		if(returnedfn)
		{
			temp_image = t3f_load_animation(returnedfn);
			if(!temp_image)
			{
				temp_image = t3f_load_animation_from_bitmap(returnedfn);
			}
			if(temp_image)
			{
				if(*app->current_image)
				{
					t3f_destroy_animation(*app->current_image);
				}
				*app->current_image = temp_image;
				crsed_mark_change(data);
			}
			strcpy(app->last_image_filename, returnedfn);
		}
	}
	return 1;
}

int crsed_menu_image_save(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	char * returnedfn;
	ALLEGRO_PATH * pp;

	if(app->course && *app->current_image)
	{
		returnedfn = ncd_file_select(1, app->last_image_filename, "Save Image File", "*.t3a");
		if(returnedfn)
		{
			pp = al_create_path(returnedfn);
			if(pp)
			{
				al_set_path_extension(pp, ".t3a");
				t3f_save_animation(*app->current_image, al_path_cstr(pp, '/'));
				strcpy(app->last_image_filename, al_path_cstr(pp, '/'));
				al_destroy_path(pp);
			}
		}
	}
	return 1;
}

int crsed_menu_image_clear(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;
	T3F_ANIMATION * temp_image;

	if(app->course)
	{
		temp_image = t3f_load_animation_from_bitmap("data/blank.png");
		if(temp_image)
		{
			*app->current_image = temp_image;
			crsed_mark_change(data);
		}
	}
	return 1;
}

int crsed_menu_music_choose(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;
	char * returnedfn = NULL;

	if(app->course)
	{
		returnedfn = ncd_file_select(0, app->last_music_filename, "Select Course Music", "*.ogg;*.xm;*.it;*.s3m;*.mod;*.wav");
		if(returnedfn)
		{
            if(app->music_enabled)
            {
                t3f_play_music(returnedfn);
            }
			strcpy(app->music_filename, returnedfn);
			strcpy(app->last_music_filename, returnedfn);
            t3f_refresh_menus();
		}
	}
	return 1;
}

int crsed_menu_music_remove(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    t3f_stop_music();
    strcpy(app->music_filename, "");
    t3f_refresh_menus();

    return 1;
}

int crsed_menu_music_listen(int id, void * data)
{
    APP_INSTANCE * app = (APP_INSTANCE *)data;

    app->music_enabled = !app->music_enabled;
    if(app->music_enabled)
    {
        if(strlen(app->music_filename))
        {
            t3f_play_music(app->music_filename);
        }
    }
    else
    {
        t3f_stop_music();
    }
    t3f_refresh_menus();

    return 1;
}

int crsed_menu_help_manual(int id, void * data)
{
	t3f_open_url("manual/index.htm");
	return 1;
}

int crsed_menu_help_about(int id, void * data)
{
	APP_INSTANCE * app = (APP_INSTANCE *)data;

	t3f_enable_menus(false);
	t3gui_show_dialog(app->about_dialog, t3f_queue, 0, data);
	return 1;
}
